"""Tests for mcp_server helper functions."""

from datetime import UTC, datetime
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

import airbyte_agent_mcp.mcp_server as mcp_mod
from airbyte_agent_mcp.mcp_server import (
    _INSTRUCTIONS,
    _TRUNCATION_SUFFIX,
    ACTION_DOWNLOAD,
    ACTION_LIST,
    ACTION_SEARCH,
    COLLECTION_ACTIONS,
    ENVELOPE_DATA_FIELD,
    MAX_TEXT_FIELD_CHARS,
    _apply_to_records,
    _compact,
    _drop_fields_from_record,
    _exclude_fields,
    _pick_fields_from_record,
    _require_connector,
    _select_fields,
    _to_dict,
    _transform_response,
    _truncate_long_text,
    current_datetime,
    get_instructions,
)

# --- _to_dict ---


class TestToDict:
    def test_converts_pydantic_model(self):
        class MyModel(BaseModel):
            id: int
            name: str

        model = MyModel(id=1, name="Alice")
        result = _to_dict(model)
        assert result == {"id": 1, "name": "Alice"}
        assert isinstance(result, dict)

    def test_converts_nested_pydantic_model(self):
        class Meta(BaseModel):
            has_more: bool

        class Envelope(BaseModel):
            data: list[dict]
            meta: Meta

        model = Envelope(data=[{"id": 1}], meta=Meta(has_more=False))
        result = _to_dict(model)
        assert result == {"data": [{"id": 1}], "meta": {"has_more": False}}
        assert isinstance(result, dict)
        assert isinstance(result["meta"], dict)

    def test_dict_passthrough(self):
        data = {"id": 1, "name": "Alice"}
        assert _to_dict(data) is data

    def test_scalar_passthrough(self):
        assert _to_dict(42) == 42
        assert _to_dict("hello") == "hello"

    def test_list_passthrough(self):
        data = [1, 2, 3]
        assert _to_dict(data) is data

    def test_select_fields_works_after_to_dict(self):
        class Record(BaseModel):
            id: int
            name: str
            bio: str

        model = Record(id=1, name="Alice", bio="long")
        result = _select_fields(_to_dict(model), ["id", "name"])
        assert result == {"id": 1, "name": "Alice"}

    def test_exclude_fields_works_after_to_dict(self):
        class Record(BaseModel):
            id: int
            name: str
            bio: str

        model = Record(id=1, name="Alice", bio="long")
        result = _exclude_fields(_to_dict(model), ["bio"])
        assert result == {"id": 1, "name": "Alice"}


# --- _compact ---


class TestCompact:
    def test_removes_none_values(self):
        assert _compact({"a": 1, "b": None}) == {"a": 1}

    def test_removes_empty_strings(self):
        assert _compact({"a": "hello", "b": ""}) == {"a": "hello"}

    def test_removes_empty_lists(self):
        assert _compact({"a": [1], "b": []}) == {"a": [1]}

    def test_removes_empty_dicts(self):
        assert _compact({"a": {"x": 1}, "b": {}}) == {"a": {"x": 1}}

    def test_nested_dict(self):
        data = {"user": {"name": "Alice", "bio": None, "tags": []}}
        assert _compact(data) == {"user": {"name": "Alice"}}

    def test_list_of_dicts(self):
        data = [{"id": 1, "x": None}, {"id": 2, "y": ""}]
        assert _compact(data) == [{"id": 1}, {"id": 2}]

    def test_preserves_falsy_but_meaningful_values(self):
        data = {"count": 0, "flag": False, "name": None}
        assert _compact(data) == {"count": 0, "flag": False}

    def test_scalar_passthrough(self):
        assert _compact(42) == 42
        assert _compact("hello") == "hello"

    def test_deeply_nested(self):
        data = {"a": {"b": {"c": None, "d": "ok"}}}
        assert _compact(data) == {"a": {"b": {"d": "ok"}}}

    def test_all_empty_collapses(self):
        data = {"a": None, "b": "", "c": [], "d": {}}
        assert _compact(data) == {}

    def test_nested_dict_empty_after_compaction(self):
        data = {"user": {"name": None, "bio": None}}
        assert _compact(data) == {}

    def test_nested_mixed_empty_after_compaction(self):
        data = {"keep": 1, "drop": {"a": None, "b": ""}}
        assert _compact(data) == {"keep": 1}


# --- _truncate_long_text ---


class TestTruncateLongText:
    def test_short_string_unchanged(self):
        assert _truncate_long_text("short") == "short"

    def test_string_at_limit_unchanged(self):
        text = "x" * MAX_TEXT_FIELD_CHARS
        assert _truncate_long_text(text) == text

    def test_string_over_limit_is_truncated(self):
        text = "x" * (MAX_TEXT_FIELD_CHARS + 100)
        result = _truncate_long_text(text)
        assert result.endswith(_TRUNCATION_SUFFIX)
        assert result.startswith("x" * MAX_TEXT_FIELD_CHARS)

    def test_non_string_unchanged(self):
        assert _truncate_long_text(42) == 42
        assert _truncate_long_text(True) is True

    def test_dict_values_truncated(self):
        data = {"id": 1, "body": "x" * 500}
        result = _truncate_long_text(data)
        assert result["id"] == 1
        assert result["body"].endswith(_TRUNCATION_SUFFIX)
        assert len(result["body"]) < 500

    def test_list_items_truncated(self):
        data = [{"text": "a" * 500}, {"text": "short"}]
        result = _truncate_long_text(data)
        assert result[0]["text"].endswith(_TRUNCATION_SUFFIX)
        assert result[1]["text"] == "short"

    def test_nested_truncation(self):
        data = {"record": {"details": {"notes": "n" * 500}}}
        result = _truncate_long_text(data)
        assert result["record"]["details"]["notes"].endswith(_TRUNCATION_SUFFIX)

    def test_preserves_structure(self):
        data = {"a": 1, "b": [2, 3], "c": {"d": "short"}}
        assert _truncate_long_text(data) == data


# --- _pick_fields_from_record ---


class TestPickFieldsFromRecord:
    def test_picks_top_level_fields(self):
        record = {"id": 1, "name": "Alice", "email": "a@b.com", "bio": "long text"}
        assert _pick_fields_from_record(record, ["id", "name"]) == {"id": 1, "name": "Alice"}

    def test_ignores_missing_fields(self):
        record = {"id": 1, "name": "Alice"}
        assert _pick_fields_from_record(record, ["id", "missing"]) == {"id": 1}

    def test_dot_notation_nested(self):
        record = {"id": 1, "content": {"topics": ["a"], "brief": "summary", "details": "long"}}
        result = _pick_fields_from_record(record, ["id", "content.topics"])
        assert result == {"id": 1, "content": {"topics": ["a"]}}

    def test_multiple_nested_from_same_parent(self):
        record = {"content": {"topics": ["a"], "brief": "b", "details": "c"}}
        result = _pick_fields_from_record(record, ["content.topics", "content.brief"])
        assert result == {"content": {"topics": ["a"], "brief": "b"}}

    def test_nested_field_parent_missing(self):
        record = {"id": 1}
        assert _pick_fields_from_record(record, ["content.topics"]) == {}

    def test_nested_field_parent_not_dict(self):
        record = {"id": 1, "content": "just a string"}
        assert _pick_fields_from_record(record, ["id", "content.topics"]) == {"id": 1}

    def test_non_dict_passthrough(self):
        assert _pick_fields_from_record("scalar", ["id"]) == "scalar"
        assert _pick_fields_from_record(42, ["id"]) == 42

    def test_empty_fields_list(self):
        record = {"id": 1, "name": "Alice"}
        assert _pick_fields_from_record(record, []) == {}

    def test_deeply_nested_dot_notation(self):
        record = {"a": {"b": {"c": "deep", "d": "other"}}}
        result = _pick_fields_from_record(record, ["a.b.c"])
        assert result == {"a": {"b": {"c": "deep"}}}


# --- _select_fields ---


class TestSelectFields:
    def test_list_of_records(self):
        data = [
            {"id": 1, "name": "Alice", "bio": "long"},
            {"id": 2, "name": "Bob", "bio": "long"},
        ]
        result = _select_fields(data, ["id", "name"])
        assert result == [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]

    def test_single_record(self):
        data = {"id": 1, "name": "Alice", "bio": "long", "email": "a@b.com"}
        result = _select_fields(data, ["id", "name"])
        assert result == {"id": 1, "name": "Alice"}

    def test_non_dict_passthrough(self):
        assert _select_fields("scalar", ["id"]) == "scalar"
        assert _select_fields(42, ["id"]) == 42

    def test_list_with_nested_fields(self):
        data = [
            {"id": 1, "content": {"topics": ["a"], "brief": "b", "details": "c"}},
        ]
        result = _select_fields(data, ["id", "content.topics"])
        assert result == [{"id": 1, "content": {"topics": ["a"]}}]

    def test_empty_list(self):
        result = _select_fields([], ["id"])
        assert result == []

    def test_search_hit_style_with_data_prefix(self):
        """Test selecting fields from search hits where entity data is nested under 'data' key."""
        hits = [
            {"data": {"id": 1, "name": "Alice", "bio": "long"}, "score": 0.9},
            {"data": {"id": 2, "name": "Bob", "bio": "long"}, "score": 0.8},
        ]
        # Using "data." prefix to select fields inside the data object
        result = _select_fields(hits, ["data.id", "data.name"])
        assert result == [
            {"data": {"id": 1, "name": "Alice"}},
            {"data": {"id": 2, "name": "Bob"}},
        ]


# --- _drop_fields_from_record ---


class TestDropFieldsFromRecord:
    def test_drops_top_level_fields(self):
        record = {"id": 1, "name": "Alice", "bio": "long text", "email": "a@b.com"}
        assert _drop_fields_from_record(record, ["bio", "email"]) == {"id": 1, "name": "Alice"}

    def test_ignores_missing_fields(self):
        record = {"id": 1, "name": "Alice"}
        assert _drop_fields_from_record(record, ["missing"]) == {"id": 1, "name": "Alice"}

    def test_dot_notation_nested(self):
        record = {"id": 1, "content": {"topics": ["a"], "brief": "summary", "details": "long"}}
        result = _drop_fields_from_record(record, ["content.brief", "content.details"])
        assert result == {"id": 1, "content": {"topics": ["a"]}}

    def test_drop_entire_nested_object(self):
        record = {"id": 1, "content": {"topics": ["a"]}, "interaction": {"stats": 42}}
        result = _drop_fields_from_record(record, ["content", "interaction"])
        assert result == {"id": 1}

    def test_nested_field_parent_missing(self):
        record = {"id": 1}
        assert _drop_fields_from_record(record, ["content.brief"]) == {"id": 1}

    def test_nested_field_parent_not_dict(self):
        record = {"id": 1, "content": "just a string"}
        assert _drop_fields_from_record(record, ["content.brief"]) == {"id": 1, "content": "just a string"}

    def test_non_dict_passthrough(self):
        assert _drop_fields_from_record("scalar", ["id"]) == "scalar"
        assert _drop_fields_from_record(42, ["id"]) == 42

    def test_empty_fields_list(self):
        record = {"id": 1, "name": "Alice"}
        assert _drop_fields_from_record(record, []) == {"id": 1, "name": "Alice"}

    def test_deeply_nested_dot_notation(self):
        record = {"a": {"b": {"c": "deep", "d": "other"}}}
        result = _drop_fields_from_record(record, ["a.b.c"])
        assert result == {"a": {"b": {"d": "other"}}}

    def test_mixed_top_and_nested(self):
        record = {"id": 1, "name": "Alice", "content": {"brief": "b", "topics": ["a"]}}
        result = _drop_fields_from_record(record, ["name", "content.brief"])
        assert result == {"id": 1, "content": {"topics": ["a"]}}


# --- _exclude_fields ---


class TestExcludeFields:
    def test_list_of_records(self):
        data = [
            {"id": 1, "name": "Alice", "bio": "long"},
            {"id": 2, "name": "Bob", "bio": "long"},
        ]
        result = _exclude_fields(data, ["bio"])
        assert result == [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]

    def test_single_record(self):
        data = {"id": 1, "name": "Alice", "bio": "long", "email": "a@b.com"}
        result = _exclude_fields(data, ["bio", "email"])
        assert result == {"id": 1, "name": "Alice"}

    def test_non_dict_passthrough(self):
        assert _exclude_fields("scalar", ["id"]) == "scalar"
        assert _exclude_fields(42, ["id"]) == 42

    def test_list_with_nested_excludes(self):
        data = [
            {"id": 1, "content": {"topics": ["a"], "brief": "b", "details": "c"}},
        ]
        result = _exclude_fields(data, ["content.brief", "content.details"])
        assert result == [{"id": 1, "content": {"topics": ["a"]}}]

    def test_empty_list(self):
        result = _exclude_fields([], ["bio"])
        assert result == []

    def test_search_hit_style_with_data_prefix(self):
        """Test excluding fields from search hits where entity data is nested under 'data' key."""
        hits = [
            {"data": {"id": 1, "name": "Alice", "bio": "long"}, "score": 0.9},
            {"data": {"id": 2, "name": "Bob", "bio": "long"}, "score": 0.8},
        ]
        # Using "data." prefix to exclude fields inside the data object
        result = _exclude_fields(hits, ["data.bio"])
        assert result == [
            {"data": {"id": 1, "name": "Alice"}, "score": 0.9},
            {"data": {"id": 2, "name": "Bob"}, "score": 0.8},
        ]


# --- _apply_to_records ---


class TestApplyToRecords:
    def test_applies_to_list_items(self):
        data = [{"x": 1}, {"x": 2}, {"x": 3}]
        result = _apply_to_records(data, lambda r: {**r, "y": r["x"] * 2})
        assert result == [{"x": 1, "y": 2}, {"x": 2, "y": 4}, {"x": 3, "y": 6}]

    def test_applies_to_single_record(self):
        data = {"x": 1}
        result = _apply_to_records(data, lambda r: {**r, "y": r["x"] * 2})
        assert result == {"x": 1, "y": 2}

    def test_empty_list(self):
        result = _apply_to_records([], lambda r: r)
        assert result == []

    def test_scalar_passthrough(self):
        result = _apply_to_records("scalar", lambda r: {"transformed": True})
        assert result == {"transformed": True}

    def test_none_passthrough(self):
        result = _apply_to_records(None, lambda r: {"transformed": True})
        assert result == {"transformed": True}


# --- Envelope preservation (simulates execute behavior) ---


class TestEnvelopePreservation:
    """Tests that simulate how execute extracts data, transforms it, and preserves the envelope."""

    def test_list_action_preserves_envelope(self):
        """For list action: extract data, transform, put back with meta preserved."""
        api_response = {
            "data": [
                {"id": 1, "name": "Alice", "bio": "long text", "extra": "stuff"},
                {"id": 2, "name": "Bob", "bio": "more text", "extra": "things"},
            ],
            "meta": {"has_more": True, "cursor": "abc123"},
        }

        # Simulate what execute does
        records = api_response.get("data", [])
        records = _select_fields(records, ["id", "name"])
        records = _truncate_long_text(records)
        records = _compact(records)
        result = {**api_response, "data": records}

        assert result == {
            "data": [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}],
            "meta": {"has_more": True, "cursor": "abc123"},
        }

    def test_search_action_preserves_envelope_and_hit_structure(self):
        """For search action: extract hits, transform data inside each hit, preserve envelope."""
        api_response = {
            "hits": [
                {"data": {"id": 1, "name": "Alice", "bio": "long"}, "score": 0.95},
                {"data": {"id": 2, "name": "Bob", "bio": "text"}, "score": 0.87},
            ],
            "meta": {"total": 100, "cursor": "xyz"},
        }

        # Simulate what execute does for search
        hits = api_response.get("hits", [])
        select_fields = ["id", "name"]
        # Execute adds "data." prefix for search
        prefixed_fields = [f"data.{f}" for f in select_fields]
        hits = _select_fields(hits, prefixed_fields)
        hits = _truncate_long_text(hits)
        hits = _compact(hits)
        result = {**api_response, "hits": hits}

        assert result == {
            "hits": [
                {"data": {"id": 1, "name": "Alice"}},
                {"data": {"id": 2, "name": "Bob"}},
            ],
            "meta": {"total": 100, "cursor": "xyz"},
        }

    def test_get_action_returns_transformed_record_directly(self):
        """For get action: no envelope, transform and return record directly."""
        api_response = {"id": 1, "name": "Alice", "bio": "long text", "email": "a@b.com"}

        # Simulate what execute does for get
        result = _select_fields(api_response, ["id", "name"])
        result = _compact(result)

        assert result == {"id": 1, "name": "Alice"}

    def test_list_action_with_exclude_fields(self):
        """For list action with exclude_fields."""
        api_response = {
            "data": [
                {"id": 1, "name": "Alice", "content": {"brief": "b", "details": "long"}},
            ],
            "meta": {"has_more": False},
        }

        records = api_response.get("data", [])
        records = _exclude_fields(records, ["content.details"])
        records = _compact(records)
        result = {**api_response, "data": records}

        assert result == {
            "data": [{"id": 1, "name": "Alice", "content": {"brief": "b"}}],
            "meta": {"has_more": False},
        }


# --- current_datetime ---


class TestCurrentDatetime:
    def test_returns_iso_format(self):
        fixed_time = datetime(2024, 6, 15, 14, 30, 45, tzinfo=UTC)
        with patch("airbyte_agent_mcp.mcp_server.datetime") as mock_dt:
            mock_dt.now.return_value = fixed_time
            result = current_datetime()

        assert result == "2024-06-15T14:30:45+00:00"

    def test_uses_utc_timezone(self):
        result = current_datetime()
        assert result.endswith("+00:00")

    def test_returns_string(self):
        result = current_datetime()
        assert isinstance(result, str)
        datetime.fromisoformat(result)


# --- get_instructions ---


class TestGetInstructions:
    def test_returns_non_empty_string(self):
        result = get_instructions()
        assert isinstance(result, str)
        assert len(result) > 0

    def test_contains_key_sections(self):
        result = get_instructions()
        assert "QUERY PLANNING" in result
        assert "ACTION SELECTION" in result
        assert "FIELD SELECTION" in result
        assert "QUERY SIZING AND PAGINATION" in result
        assert "DATE RANGE QUERIES" in result
        assert "DOWNLOADS" in result

    def test_returns_same_as_constant(self):
        result = get_instructions()
        assert result == _INSTRUCTIONS


# --- _require_connector ---


class TestRequireConnector:
    def test_returns_connector_when_initialized(self):
        mock_connector = MagicMock()
        original = mcp_mod._connector
        mcp_mod._connector = mock_connector
        try:
            result = _require_connector()
            assert result is mock_connector
        finally:
            mcp_mod._connector = original

    def test_raises_when_not_initialized(self):
        original = mcp_mod._connector
        mcp_mod._connector = None
        try:
            with pytest.raises(RuntimeError, match="Connector not initialized"):
                _require_connector()
        finally:
            mcp_mod._connector = original


# --- _transform_response ---


class TestTransformResponse:
    def test_list_action_extracts_and_replaces_data(self):
        result = {"data": [{"id": 1, "name": "Alice"}], "meta": {"has_more": False}}
        transformed = _transform_response(result, ACTION_LIST, None, None)
        assert ENVELOPE_DATA_FIELD[ACTION_LIST] in transformed
        assert transformed["data"] == [{"id": 1, "name": "Alice"}]
        assert transformed["meta"] == {"has_more": False}

    def test_search_action_extracts_and_replaces_hits(self):
        result = {"hits": [{"data": {"id": 1, "name": "Alice"}}], "meta": {"total": 1}}
        transformed = _transform_response(result, ACTION_SEARCH, None, None)
        assert ENVELOPE_DATA_FIELD[ACTION_SEARCH] in transformed
        assert transformed["hits"] == [{"data": {"id": 1, "name": "Alice"}}]
        assert transformed["meta"] == {"total": 1}

    def test_search_prefixes_select_fields_with_data(self):
        result = {
            "hits": [{"data": {"id": 1, "name": "Alice", "bio": "long"}, "score": 0.9}],
            "meta": {},
        }
        transformed = _transform_response(result, ACTION_SEARCH, ["id", "name"], None)
        assert transformed["hits"] == [{"data": {"id": 1, "name": "Alice"}}]

    def test_search_prefixes_exclude_fields_with_data(self):
        result = {
            "hits": [{"data": {"id": 1, "name": "Alice", "bio": "long"}, "score": 0.9}],
            "meta": {},
        }
        transformed = _transform_response(result, ACTION_SEARCH, None, ["bio"])
        assert transformed["hits"] == [{"data": {"id": 1, "name": "Alice"}, "score": 0.9}]

    def test_get_action_returns_transformed_directly(self):
        result = {"id": 1, "name": "Alice", "bio": "long"}
        transformed = _transform_response(result, "get", ["id", "name"], None)
        assert transformed == {"id": 1, "name": "Alice"}

    def test_applies_select_fields_for_list(self):
        result = {"data": [{"id": 1, "name": "Alice", "bio": "long"}], "meta": {}}
        transformed = _transform_response(result, ACTION_LIST, ["id", "name"], None)
        assert transformed["data"] == [{"id": 1, "name": "Alice"}]

    def test_applies_exclude_fields_for_list(self):
        result = {"data": [{"id": 1, "name": "Alice", "bio": "long"}], "meta": {}}
        transformed = _transform_response(result, ACTION_LIST, None, ["bio"])
        assert transformed["data"] == [{"id": 1, "name": "Alice"}]

    def test_select_fields_takes_priority(self):
        result = {"data": [{"id": 1, "name": "Alice", "bio": "long"}], "meta": {}}
        transformed = _transform_response(result, ACTION_LIST, ["id"], ["name"])
        assert transformed["data"] == [{"id": 1}]

    def test_truncates_for_list(self):
        long_text = "x" * 500
        result = {"data": [{"id": 1, "text": long_text}], "meta": {}}
        transformed = _transform_response(result, ACTION_LIST, None, None)
        assert transformed["data"][0]["text"].endswith(_TRUNCATION_SUFFIX)
        assert len(transformed["data"][0]["text"]) < 500

    def test_truncates_for_search(self):
        long_text = "x" * 500
        result = {"hits": [{"data": {"id": 1, "text": long_text}}], "meta": {}}
        transformed = _transform_response(result, ACTION_SEARCH, None, None)
        assert transformed["hits"][0]["data"]["text"].endswith(_TRUNCATION_SUFFIX)

    def test_does_not_truncate_for_get(self):
        long_text = "x" * 500
        result = {"id": 1, "text": long_text}
        transformed = _transform_response(result, "get", None, None)
        assert transformed["text"] == long_text

    def test_compacts_result(self):
        result = {"data": [{"id": 1, "empty": None, "blank": ""}], "meta": {}}
        transformed = _transform_response(result, ACTION_LIST, None, None)
        assert transformed["data"] == [{"id": 1}]

    def test_preserves_envelope_meta(self):
        result = {"data": [{"id": 1}], "meta": {"cursor": "abc", "has_more": True}}
        transformed = _transform_response(result, ACTION_LIST, None, None)
        assert transformed["meta"] == {"cursor": "abc", "has_more": True}

    def test_list_missing_data_envelope_raises(self):
        result = {"items": [{"id": 1}], "meta": {}}
        with pytest.raises(KeyError, match="List response missing 'data' envelope key"):
            _transform_response(result, ACTION_LIST, None, None)

    def test_search_missing_hits_envelope_raises(self):
        result = {"results": [{"id": 1}], "meta": {}}
        with pytest.raises(KeyError, match="Search response missing 'hits' envelope key"):
            _transform_response(result, ACTION_SEARCH, None, None)


# --- _get_save_download ---


class TestGetSaveDownload:
    def test_imports_from_connector_package(self):
        mock_connector = MagicMock()
        mock_connector.__module__ = "airbyte_agent_gong.connector"
        type(mock_connector).__module__ = "airbyte_agent_gong.connector"

        mock_module = MagicMock()
        mock_module.save_download = lambda: "saved"

        original = mcp_mod._connector
        mcp_mod._connector = mock_connector
        try:
            with patch("importlib.import_module", return_value=mock_module) as mock_import:
                from airbyte_agent_mcp.mcp_server import _get_save_download

                result = _get_save_download()
                mock_import.assert_called_once_with("airbyte_agent_gong._vendored.connector_sdk")
                assert result == mock_module.save_download
        finally:
            mcp_mod._connector = original

    def test_returns_save_download_function(self):
        mock_connector = MagicMock()
        type(mock_connector).__module__ = "airbyte_agent_test.connector"

        mock_save_download = MagicMock()
        mock_module = MagicMock()
        mock_module.save_download = mock_save_download

        original = mcp_mod._connector
        mcp_mod._connector = mock_connector
        try:
            with patch("importlib.import_module", return_value=mock_module):
                from airbyte_agent_mcp.mcp_server import _get_save_download

                result = _get_save_download()
                assert result is mock_save_download
        finally:
            mcp_mod._connector = original


# --- Constants ---


class TestConstants:
    def test_action_constants_are_strings(self):
        assert ACTION_LIST == "list"
        assert ACTION_SEARCH == "search"
        assert ACTION_DOWNLOAD == "download"

    def test_collection_actions_contains_list_and_search(self):
        assert ACTION_LIST in COLLECTION_ACTIONS
        assert ACTION_SEARCH in COLLECTION_ACTIONS
        assert ACTION_DOWNLOAD not in COLLECTION_ACTIONS

    def test_envelope_data_field_mapping(self):
        assert ENVELOPE_DATA_FIELD[ACTION_LIST] == "data"
        assert ENVELOPE_DATA_FIELD[ACTION_SEARCH] == "hits"



"""Tests for login CLI command."""

from unittest.mock import patch

from typer.testing import CliRunner

from airbyte_agent_mcp.cli import cli

runner = CliRunner()


def _invoke_login(tmp_path, org_id="my-org", input_text="my-id\nmy-secret\n"):
    with (
        patch("airbyte_agent_mcp.cli.login_commands.get_org_env_path", return_value=tmp_path / "orgs" / org_id / ".env"),
        patch("airbyte_agent_mcp.cli.login_commands.Config") as mock_config_cls,
    ):
        mock_config = mock_config_cls.load.return_value
        result = runner.invoke(cli, ["login", org_id], input=input_text)
    return result, mock_config


class TestLogin:
    def test_writes_env_file_to_org_dir(self, tmp_path):
        result, _ = _invoke_login(tmp_path)

        assert result.exit_code == 0
        env_file = tmp_path / "orgs" / "my-org" / ".env"
        assert env_file.exists()

        content = env_file.read_text()
        assert "AIRBYTE_CLIENT_ID=my-id" in content
        assert "AIRBYTE_CLIENT_SECRET=my-secret" in content
        assert "AIRBYTE_ORGANIZATION_ID=my-org" in content

    def test_sets_default_organization(self, tmp_path):
        _, mock_config = _invoke_login(tmp_path)

        assert mock_config.default_organization_id == "my-org"
        mock_config.save.assert_called_once()

    def test_prints_confirmation(self, tmp_path):
        result, _ = _invoke_login(tmp_path)

        assert result.exit_code == 0
        assert "Credentials saved" in result.output

    def test_prints_default_org_message(self, tmp_path):
        result, _ = _invoke_login(tmp_path)

        assert result.exit_code == 0
        assert "set as default" in result.output

    def test_shows_credentials_url(self, tmp_path):
        result, _ = _invoke_login(tmp_path, org_id="org-abc")

        assert result.exit_code == 0
        assert "https://app.airbyte.ai/organizations/org-abc/authentication-module" in result.output

    def test_overwrites_existing_file(self, tmp_path):
        org_dir = tmp_path / "orgs" / "my-org"
        org_dir.mkdir(parents=True)
        env_file = org_dir / ".env"
        env_file.write_text("AIRBYTE_CLIENT_ID=old-id\n")

        result, _ = _invoke_login(tmp_path, input_text="new-id\nnew-secret\n")

        assert result.exit_code == 0
        content = env_file.read_text()
        assert "AIRBYTE_CLIENT_ID=new-id" in content
        assert "old-id" not in content

    def test_creates_org_directory(self, tmp_path):
        nested = tmp_path / "sub" / "dir"
        with (
            patch("airbyte_agent_mcp.cli.login_commands.get_org_env_path", return_value=nested / "orgs" / "org-1" / ".env"),
            patch("airbyte_agent_mcp.cli.login_commands.Config"),
        ):
            result = runner.invoke(cli, ["login", "org-1"], input="id\nsecret\n")

        assert result.exit_code == 0
        assert (nested / "orgs" / "org-1" / ".env").exists()



"""Shared test fixtures for CLI tests."""

from unittest.mock import MagicMock

import pytest
from typer.testing import CliRunner

from airbyte_agent_mcp.cli import cli


@pytest.fixture
def cli_runner():
    return CliRunner()


@pytest.fixture
def mock_connector():
    connector = MagicMock()
    connector.connector_name = "test-connector"
    connector.connector_version = "1.0.0"
    return connector


@pytest.fixture
def invoke(cli_runner):
    """Helper to invoke CLI commands."""

    def _invoke(*args, **kwargs):
        return cli_runner.invoke(cli, list(args), **kwargs)

    return _invoke



"""Connector configuration models."""

import os
import re
from pathlib import Path

import yaml
from pydantic import BaseModel, Field, model_validator

_ENV_PATTERN = re.compile(r"\$\{env\.([^}]+)\}")


def resolve_env_vars(value: str) -> str:
    """Replace ${env.VAR_NAME} placeholders with actual environment variable values.

    Args:
        value: String that may contain ${env.VAR_NAME} placeholders.

    Returns:
        String with placeholders replaced by environment variable values.

    Raises:
        ValueError: If an environment variable is not set.
    """

    def replace_env(match: re.Match) -> str:
        var_name = match.group(1)
        env_value = os.environ.get(var_name)
        if env_value is None:
            raise ValueError(f"Environment variable '{var_name}' is not set")
        return env_value

    return _ENV_PATTERN.sub(replace_env, value)


def resolve_credentials(credentials: dict[str, str]) -> dict[str, str]:
    """Resolve environment variables in all credential values.

    Args:
        credentials: Dict with potentially templated values.

    Returns:
        Dict with all environment variable placeholders resolved.

    Raises:
        ValueError: If any environment variable is not set.
    """
    return {key: resolve_env_vars(value) for key, value in credentials.items()}


class ConnectorSource(BaseModel):
    """Connector source configuration.

    Supports three mutually exclusive package source types:
    - ``package``: A PyPI package name (e.g., ``airbyte-agent-gong``)
    - ``path``: A local filesystem path (e.g., ``../integrations/gong/.generated``)
    - ``git``: A git repository URL (e.g., ``https://github.com/org/repo.git``)

    Additionally, ``connector_id`` is orthogonal and can be combined with any
    package source, or used alone (auto-resolves the package from the registry).
    """

    package: str | None = None
    version: str | None = None
    path: str | None = None
    git: str | None = None
    ref: str | None = None
    subdirectory: str | None = None
    connector_id: str | None = None

    @model_validator(mode="after")
    def validate_source(self) -> "ConnectorSource":
        sources = [s for s in (self.package, self.path, self.git) if s is not None]
        if len(sources) > 1:
            raise ValueError("At most one of 'package', 'path', 'git' may be specified")
        if len(sources) == 0 and self.connector_id is None:
            raise ValueError("At least one of 'package', 'path', 'git', or 'connector_id' must be specified")
        if self.version is not None and self.package is None:
            raise ValueError("'version' can only be used with 'package'")
        if self.ref is not None and self.git is None:
            raise ValueError("'ref' can only be used with 'git'")
        if self.subdirectory is not None and self.git is None:
            raise ValueError("'subdirectory' can only be used with 'git'")
        return self

    @property
    def is_cloud(self) -> bool:
        return self.connector_id is not None

    @property
    def has_package_source(self) -> bool:
        return self.package is not None or self.path is not None or self.git is not None

    def to_install_spec(self) -> str | None:
        """Return a pip-installable spec string, or None if cloud-only."""
        if self.path is not None:
            return self.path
        if self.git is not None:
            spec = f"git+{self.git}"
            if self.ref:
                spec += f"@{self.ref}"
            if self.subdirectory:
                spec += f"#subdirectory={self.subdirectory}"
            return spec
        if self.package is not None:
            if self.version:
                return f"{self.package}=={self.version}"
            return self.package
        return None


def auth_config_to_env_template(auth_config_type: type[BaseModel], connector_name: str | None = None) -> dict[str, str]:
    """Convert an auth config type to a dict with environment variable placeholders.

    Args:
        auth_config_type: A Pydantic model class for auth configuration.
        connector_name: Optional connector name used as env var prefix (e.g., "gong").
            When None, field names are used directly (e.g., ``airbyte_client_id`` -> ``AIRBYTE_CLIENT_ID``).

    Returns:
        Dict mapping field names to environment variable placeholders.
    """
    if connector_name:
        prefix = connector_name.upper().replace("-", "_").replace(" ", "_")
        return {field_name: f"${{env.{prefix}_{field_name.upper()}}}" for field_name in auth_config_type.model_fields}
    return {field_name: f"${{env.{field_name.upper()}}}" for field_name in auth_config_type.model_fields}


def connector_params_to_template(params: list) -> dict[str, str]:
    """Convert a list of ConnectorParam to a dict with placeholder values.

    Args:
        params: List of ConnectorParam from get_additional_connector_params().

    Returns:
        Dict mapping param names to placeholder values.
        E.g., {"subdomain": "# TODO: your-subdomain"}
    """
    return {p.name: f"# TODO: {p.name}" for p in params}


_PACKAGE_PREFIX = "airbyte-agent-"


class ConnectorConfig(BaseModel):
    """Configuration for a single connector."""

    connector: ConnectorSource
    credentials: dict[str, str] = Field(default_factory=dict)
    config: dict[str, str] = Field(default_factory=dict)

    @model_validator(mode="before")
    @classmethod
    def _normalize_connector(cls, data: dict) -> dict:
        if not isinstance(data, dict) or "connector" not in data:
            return data
        value = data["connector"]
        if isinstance(value, str):
            name = value if value.startswith(_PACKAGE_PREFIX) else f"{_PACKAGE_PREFIX}{value}"
            data["connector"] = {"package": name}
        return data

    @classmethod
    def load(cls, config_path: Path) -> "ConnectorConfig":
        """Load connector configuration from file.

        Args:
            config_path: Path to config file.

        Returns:
            ConnectorConfig object with loaded configuration.

        Raises:
            FileNotFoundError: If config file doesn't exist.
        """
        if not config_path.exists():
            raise FileNotFoundError(f"Connector config file not found: {config_path}")

        with open(config_path) as f:
            data = yaml.safe_load(f) or {}
            return cls(**data)

    def to_dict(self) -> dict:
        """Convert to dict, excluding None values and empty dicts.

        Returns:
            Dict representation with empty values removed.
        """
        data = self.model_dump(mode="json", exclude_none=True)
        return {k: v for k, v in data.items() if v != {}}

    def save(self, config_path: Path) -> None:
        """Save connector configuration to file.

        Args:
            config_path: Path to config file.
        """
        config_path.parent.mkdir(parents=True, exist_ok=True)

        with open(config_path, "w") as f:
            yaml.safe_dump(self.to_dict(), f, default_flow_style=False, sort_keys=False)



"""Tests for installer module."""

import importlib.metadata
import json
import subprocess
from unittest.mock import MagicMock, patch

import pytest

from airbyte_agent_mcp.installer import (
    ConnectorInstallError,
    _find_package_by_source_url,
    _get_pip_command,
    get_package_name,
    install_package,
    is_git_url,
    is_local_path,
    is_package_installed,
    uninstall_package,
)


class TestIsLocalPath:
    def test_absolute_path(self):
        assert is_local_path("/path/to/connector") is True

    def test_relative_path_dot_slash(self):
        assert is_local_path("./my/connector") is True

    def test_relative_path_no_dot(self):
        assert is_local_path("myconnectors/gong") is True

    def test_home_path(self):
        assert is_local_path("~/connectors/gong") is True

    def test_parent_path(self):
        assert is_local_path("../other/connector") is True

    def test_pypi_name_not_path(self):
        assert is_local_path("airbyte-agent-gong") is False

    def test_pypi_name_with_version_not_path(self):
        assert is_local_path("airbyte-agent-gong==1.0.0") is False

    def test_git_url_not_path(self):
        assert is_local_path("git+https://github.com/org/repo.git") is False


class TestIsGitUrl:
    def test_git_https(self):
        assert is_git_url("git+https://github.com/org/repo.git") is True

    def test_git_with_branch(self):
        assert is_git_url("git+https://github.com/org/repo.git@main") is True

    def test_git_with_subdirectory(self):
        assert is_git_url("git+https://github.com/org/repo.git#subdirectory=pkg") is True

    def test_pypi_name_not_git(self):
        assert is_git_url("airbyte-agent-gong") is False

    def test_local_path_not_git(self):
        assert is_git_url("/path/to/connector") is False

    def test_https_url_without_git_prefix(self):
        assert is_git_url("https://github.com/org/repo.git") is False


class TestGetPipCommand:
    @patch("airbyte_agent_mcp.installer.is_executable_available", return_value=True)
    def test_uses_uv_when_available(self, _mock):
        assert _get_pip_command() == ["uv", "pip"]

    @patch("airbyte_agent_mcp.installer.is_executable_available", return_value=False)
    def test_falls_back_to_python_pip(self, _mock):
        cmd = _get_pip_command()
        assert cmd[1:] == ["-m", "pip"]


class TestGetPackageName:
    def test_pypi_name(self):
        assert get_package_name("airbyte-agent-gong") == "airbyte-agent-gong"

    def test_pypi_name_with_version(self):
        assert get_package_name("airbyte-agent-gong==1.0.0") == "airbyte-agent-gong"

    def test_local_path(self, tmp_path):
        pyproject = tmp_path / "pyproject.toml"
        pyproject.write_text('[project]\nname = "my-connector"\n')
        assert get_package_name(str(tmp_path)) == "my-connector"

    def test_local_path_missing_pyproject(self, tmp_path):
        with pytest.raises(ConnectorInstallError, match="pyproject.toml not found"):
            get_package_name(str(tmp_path / "nonexistent"))

    def test_git_url_found(self):
        mock_dist = MagicMock()
        mock_dist.read_text.return_value = json.dumps({"url": "https://github.com/org/repo.git"})
        mock_dist.metadata = {"Name": "my-connector"}

        with patch.object(importlib.metadata, "distributions", return_value=[mock_dist]):
            assert get_package_name("git+https://github.com/org/repo.git") == "my-connector"

    def test_git_url_not_found(self):
        with patch.object(importlib.metadata, "distributions", return_value=[]):
            with pytest.raises(ConnectorInstallError, match="Could not determine package name"):
                get_package_name("git+https://github.com/org/repo.git")


class TestInstallPackage:
    @patch("airbyte_agent_mcp.installer.subprocess.run")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    def test_installs_pypi_package(self, _pip, mock_run):
        install_package("airbyte-agent-gong")
        args = mock_run.call_args[0][0]
        assert args == ["uv", "pip", "install", "airbyte-agent-gong"]

    @patch("airbyte_agent_mcp.installer.subprocess.run")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    def test_installs_with_version(self, _pip, mock_run):
        install_package("airbyte-agent-gong==1.0.0")
        args = mock_run.call_args[0][0]
        assert args == ["uv", "pip", "install", "airbyte-agent-gong==1.0.0"]

    @patch("airbyte_agent_mcp.installer.subprocess.run")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    def test_installs_git_url(self, _pip, mock_run):
        install_package("git+https://github.com/org/repo.git@main")
        args = mock_run.call_args[0][0]
        assert args == ["uv", "pip", "install", "git+https://github.com/org/repo.git@main"]

    @patch("airbyte_agent_mcp.installer.subprocess.run")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    def test_installs_local_path_as_editable(self, _pip, mock_run, tmp_path):
        local_dir = tmp_path / "connector"
        local_dir.mkdir()
        install_package(str(local_dir))
        args = mock_run.call_args[0][0]
        assert "-e" in args
        assert str(local_dir) in args

    def test_raises_for_missing_local_path(self):
        with pytest.raises(ConnectorInstallError, match="does not exist"):
            install_package("/nonexistent/path")

    def test_raises_for_file_path(self, tmp_path):
        f = tmp_path / "file.txt"
        f.write_text("hi")
        with pytest.raises(ConnectorInstallError, match="not a directory"):
            install_package(str(f))

    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    @patch("airbyte_agent_mcp.installer.subprocess.run", side_effect=subprocess.CalledProcessError(1, "cmd", stderr="err"))
    def test_raises_on_failure(self, _run, _pip):
        with pytest.raises(ConnectorInstallError, match="Failed to install"):
            install_package("bad-package")

    @patch("site.main")
    @patch("importlib.invalidate_caches")
    @patch("airbyte_agent_mcp.installer.subprocess.run")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    def test_refreshes_import_caches_after_install(self, _pip, _run, mock_invalidate, mock_site_main):
        install_package("airbyte-agent-gong")
        mock_invalidate.assert_called_once()
        mock_site_main.assert_called_once()

    @patch("site.main")
    @patch("importlib.invalidate_caches")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    @patch("airbyte_agent_mcp.installer.subprocess.run", side_effect=subprocess.CalledProcessError(1, "cmd", stderr="err"))
    def test_does_not_refresh_caches_on_failure(self, _run, _pip, mock_invalidate, mock_site_main):
        with pytest.raises(ConnectorInstallError):
            install_package("bad-package")
        mock_invalidate.assert_not_called()
        mock_site_main.assert_not_called()


class TestFindPackageBySourceUrl:
    def test_finds_matching_package(self):
        mock_dist = MagicMock()
        mock_dist.read_text.return_value = json.dumps({"url": "https://github.com/org/repo.git", "vcs_info": {"vcs": "git"}})
        mock_dist.metadata = {"Name": "my-connector"}

        with patch.object(importlib.metadata, "distributions", return_value=[mock_dist]):
            result = _find_package_by_source_url("git+https://github.com/org/repo.git")
            assert result == "my-connector"

    def test_returns_none_when_no_match(self):
        mock_dist = MagicMock()
        mock_dist.read_text.return_value = json.dumps({"url": "https://other.com/pkg", "vcs_info": {"vcs": "git"}})
        mock_dist.metadata = {"Name": "other-package"}

        with patch.object(importlib.metadata, "distributions", return_value=[mock_dist]):
            result = _find_package_by_source_url("git+https://github.com/org/repo.git")
            assert result is None

    def test_returns_none_when_no_direct_url(self):
        mock_dist = MagicMock()
        mock_dist.read_text.return_value = None

        with patch.object(importlib.metadata, "distributions", return_value=[mock_dist]):
            result = _find_package_by_source_url("git+https://github.com/org/repo.git")
            assert result is None

    def test_returns_none_when_no_distributions(self):
        with patch.object(importlib.metadata, "distributions", return_value=[]):
            result = _find_package_by_source_url("git+https://github.com/org/repo.git")
            assert result is None

    def test_strips_git_plus_prefix_for_matching(self):
        mock_dist = MagicMock()
        mock_dist.read_text.return_value = json.dumps({"url": "https://github.com/org/repo.git"})
        mock_dist.metadata = {"Name": "my-connector"}

        with patch.object(importlib.metadata, "distributions", return_value=[mock_dist]):
            result = _find_package_by_source_url("git+https://github.com/org/repo.git@v1.0")
            assert result == "my-connector"


class TestUninstallPackage:
    @patch("airbyte_agent_mcp.installer.subprocess.run")
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    def test_uninstalls(self, _pip, mock_run):
        uninstall_package("airbyte-agent-gong")
        args = mock_run.call_args[0][0]
        assert args == ["uv", "pip", "uninstall", "-y", "airbyte-agent-gong"]

    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    @patch("airbyte_agent_mcp.installer.subprocess.run", side_effect=subprocess.CalledProcessError(1, "cmd", stderr="err"))
    def test_raises_on_failure(self, _run, _pip):
        with pytest.raises(ConnectorInstallError, match="Failed to uninstall"):
            uninstall_package("bad-package")


class TestIsPackageInstalled:
    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    @patch("airbyte_agent_mcp.installer.subprocess.run")
    def test_returns_true_when_installed(self, mock_run, _pip):
        mock_run.return_value.returncode = 0
        assert is_package_installed("airbyte-agent-gong") is True

    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    @patch("airbyte_agent_mcp.installer.subprocess.run")
    def test_returns_false_when_not_installed(self, mock_run, _pip):
        mock_run.return_value.returncode = 1
        assert is_package_installed("nonexistent") is False

    @patch("airbyte_agent_mcp.installer._get_pip_command", return_value=["uv", "pip"])
    @patch("airbyte_agent_mcp.installer.subprocess.run", side_effect=Exception("boom"))
    def test_returns_false_on_exception(self, _run, _pip):
        assert is_package_installed("anything") is False



"""Pydantic models for configuration."""

from .cli_config import Config
from .connector_config import (
    ConnectorConfig,
    ConnectorSource,
)

__all__ = [
    "Config",
    "ConnectorConfig",
    "ConnectorSource",
]



"""Shared CLI helper functions."""

from pathlib import Path

import typer
from rich.console import Console

from ..connector_utils import ConnectorLoadError, load_connector
from ..models.connector_config import ConnectorConfig


def load_connector_from_config(config: Path, console: Console):
    """Load a connector from a YAML config file.

    Handles FileNotFoundError, ConnectorLoadError, and ValueError
    by printing an error message and raising typer.Exit(1).
    """
    try:
        connector_config = ConnectorConfig.load(config)
        console.print(f"[dim]Loading connector from config: {config}[/dim]")
        connector = load_connector(connector_config)
        console.print(f"[green]\u2713[/green] Loaded connector: {connector.connector_name} v{connector.connector_version}")
        return connector
    except FileNotFoundError as e:
        console.print(f"[red]Error: {e}[/red]")
        raise typer.Exit(1) from None
    except ConnectorLoadError as e:
        console.print(f"[red]Connector load error: {e}[/red]")
        raise typer.Exit(1) from None
    except ValueError as e:
        console.print(f"[red]Configuration error: {e}[/red]")
        raise typer.Exit(1) from None



"""Utilities for exploring and introspecting connector packages."""

from __future__ import annotations

import importlib
import inspect
import typing
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel, Field

from .airbyte_api import AirbyteApi, fetch_registry
from .installer import get_package_name, install_package
from .models.connector_config import (
    ConnectorConfig,
    ConnectorSource,
    resolve_credentials,
)


@dataclass
class ConnectorParam:
    """A connector parameter definition."""

    name: str
    annotation: str | None
    required: bool


# Parameters that are always present in connector __init__ and should be excluded
_STANDARD_CONNECTOR_PARAMS = frozenset(
    {
        "self",
        "auth_config",
        "on_token_refresh",
    }
)


class AirbyteCloudAuthConfig(BaseModel):
    """Authentication configuration for Airbyte Cloud hosted execution."""

    airbyte_client_id: str = Field(description="Airbyte OAuth client ID")
    airbyte_client_secret: str = Field(description="Airbyte OAuth client secret")


def _package_to_module_name(package_name: str) -> str:
    """Convert a package name to a Python module name."""
    return package_name.replace("-", "_")


def _load_connector_module(package_name: str) -> Any:
    """Load a connector module by package name."""
    module_name = _package_to_module_name(package_name)
    return importlib.import_module(module_name)


def get_cloud_auth_config_type() -> type[BaseModel]:
    """Get the AuthConfig type for Airbyte Cloud hosted execution.

    Returns:
        The AirbyteCloudAuthConfig model class.
    """
    return AirbyteCloudAuthConfig


def _get_airbyte_auth_config_class(package_name: str) -> type[BaseModel] | None:
    """Get the hosted auth config class from a connector module.

    Uses the public ``AirbyteAuthConfig`` alias exported by the connector
    package rather than reaching into the vendored SDK internals.

    Returns:
        The hosted auth config class, or None if not found.
    """
    module = _load_connector_module(package_name)
    return getattr(module, "AirbyteAuthConfig", None)


def get_auth_config_types(package_name: str) -> list[type[BaseModel]]:
    """Get the AuthConfig types from a connector package.

    Args:
        package_name: PyPI package name (e.g., "airbyte-agent-gong").

    Returns:
        List of Pydantic model classes for each auth method.

    Raises:
        ModuleNotFoundError: If the package is not installed.
        ValueError: If the module doesn't have an AuthConfig export.
    """
    module = _load_connector_module(package_name)
    module_name = _package_to_module_name(package_name)

    # Skip AirbyteAuthConfig / AirbyteHostedAuthConfig — that's the hosted mode auth,
    # not connector-specific auth config.
    hosted_class = _get_airbyte_auth_config_class(package_name)
    auth_config_names = [name for name in dir(module) if name.endswith("AuthConfig") and getattr(module, name) is not hosted_class]
    if not auth_config_names:
        raise ValueError(f"No AuthConfig type found in {module_name}")

    auth_type = getattr(module, auth_config_names[0])
    auth_options = typing.get_args(auth_type)

    if not auth_options:
        # Not a Union, single auth type
        return [auth_type]

    return list(auth_options)


def _get_connector_class(package_name: str) -> type:
    """Get the Connector class from a package."""
    module = _load_connector_module(package_name)
    connector_names = [name for name in dir(module) if name.endswith("Connector")]
    if not connector_names:
        module_name = _package_to_module_name(package_name)
        raise ValueError(f"No Connector class found in {module_name}")
    return getattr(module, connector_names[0])


def get_additional_connector_params(package_name: str) -> list[ConnectorParam]:
    """Get additional (non-auth) parameters for a connector.

    Args:
        package_name: PyPI package name (e.g., "airbyte-agent-zendesk-support").

    Returns:
        List of ConnectorParam for parameters beyond the standard auth params.

    Raises:
        ModuleNotFoundError: If the package is not installed.
        ValueError: If the module doesn't have a Connector class.
    """
    connector_class = _get_connector_class(package_name)
    sig = inspect.signature(connector_class.__init__)

    params = []
    for name, param in sig.parameters.items():
        if name in _STANDARD_CONNECTOR_PARAMS:
            continue

        # Get the type annotation as string
        annotation = None
        if param.annotation != inspect.Parameter.empty:
            ann = param.annotation
            # Convert to string, stripping Optional/None for cleaner output
            if isinstance(ann, str):
                # Already a string annotation (forward reference)
                annotation = ann.replace(" | None", "").replace("None | ", "")
            else:
                # Actual type - get string representation
                annotation = getattr(ann, "__name__", str(ann))

        # Determine if required (no default value)
        required = param.default == inspect.Parameter.empty

        params.append(ConnectorParam(name=name, annotation=annotation, required=required))

    return params


class ConnectorLoadError(Exception):
    """Exception raised when connector loading fails."""

    pass


def _install_and_get_package_name(source: ConnectorSource) -> str:
    """Install a connector package from a ConnectorSource and return the package name.

    Args:
        source: ConnectorSource with package/path/git info.

    Returns:
        The installed Python package name.

    Raises:
        ConnectorLoadError: If installation fails.
    """
    spec = source.to_install_spec()
    if spec is None:
        raise ConnectorLoadError("No package source specified (connector_id-only requires API lookup)")
    install_package(spec)
    return get_package_name(spec)


def _resolve_cloud_package_name(source: ConnectorSource, resolved_credentials: dict[str, str]) -> str:
    """Resolve package name for a cloud-only source via Airbyte API.

    Args:
        source: ConnectorSource with connector_id.
        resolved_credentials: Already-resolved credentials dict.

    Returns:
        The PyPI package name (e.g., "airbyte-agent-gong").

    Raises:
        ConnectorLoadError: If API lookup fails.
    """
    required_cred_keys = list(AirbyteCloudAuthConfig.model_fields)
    missing_cred_keys = [k for k in required_cred_keys if k not in resolved_credentials]
    if missing_cred_keys:
        raise ConnectorLoadError(f"Cloud source requires credentials: {required_cred_keys}. Missing: {missing_cred_keys}")

    if source.connector_id is None:
        raise ConnectorLoadError("connector_id is required for cloud source resolution")

    try:
        api = AirbyteApi(
            client_id=resolved_credentials["airbyte_client_id"],
            client_secret=resolved_credentials["airbyte_client_secret"],
        )
        connector_source = api.get_connector(source.connector_id)
    except Exception as e:
        raise ConnectorLoadError(f"Failed to fetch connector source: {e}") from e

    defn_id = connector_source.get("source_template", {}).get("source_definition_id", "")

    try:
        registry_entries = fetch_registry()
    except Exception as e:
        raise ConnectorLoadError(f"Failed to fetch connector registry: {e}") from e

    registry = {e["connector_definition_id"]: e for e in registry_entries if "connector_definition_id" in e}
    reg = registry.get(defn_id, {})
    connector_name = reg.get("connector_name")
    if not connector_name:
        raise ConnectorLoadError(f"Could not determine connector type for ID: {source.connector_id}")

    return f"airbyte-agent-{connector_name}"


def get_connector_name(config: ConnectorConfig) -> str:
    """Get the connector name by installing the package and reading the class attribute.

    This does NOT instantiate the connector (no credentials needed).

    Args:
        config: ConnectorConfig with source info.

    Returns:
        The connector name (e.g. "gong", "zendesk-support").

    Raises:
        ConnectorLoadError: If the connector package cannot be installed or loaded.
    """
    source = config.connector

    if not source.has_package_source:
        raise ConnectorLoadError("Cannot determine connector name from cloud-only source without credentials")

    package_name = _install_and_get_package_name(source)

    try:
        connector_class = _get_connector_class(package_name)
    except ValueError as e:
        raise ConnectorLoadError(str(e)) from e

    name = getattr(connector_class, "connector_name", None)
    if not name:
        raise ConnectorLoadError(f"Connector class {connector_class.__name__} has no connector_name attribute")
    return name


def _find_matching_auth_config(package_name: str, credentials: dict[str, str]) -> BaseModel:
    """Find the auth config type that matches the provided credentials.

    Args:
        package_name: PyPI package name.
        credentials: Resolved credentials dict.

    Returns:
        Instantiated auth config object.

    Raises:
        ConnectorLoadError: If no matching auth config is found.
    """
    auth_config_types = get_auth_config_types(package_name)

    # Try each auth config type
    for auth_class in auth_config_types:
        try:
            return auth_class(**credentials)
        except Exception:
            continue

    auth_names = [t.__name__ for t in auth_config_types]
    raise ConnectorLoadError(f"Credentials do not match any AuthConfig in {package_name}. Available configs: {auth_names}")


def load_connector(config: ConnectorConfig) -> Any:
    """Load and instantiate a connector from configuration.

    Args:
        config: ConnectorConfig with source, credentials, and config.

    Returns:
        Instantiated connector object.

    Raises:
        ConnectorLoadError: If connector cannot be loaded or instantiated.
    """
    source = config.connector

    # Resolve environment variables in credentials
    try:
        resolved_credentials = resolve_credentials(config.credentials)
    except ValueError as e:
        raise ConnectorLoadError(str(e)) from e

    # Determine package name and install
    if source.has_package_source:
        package_name = _install_and_get_package_name(source)
    elif source.is_cloud:
        package_name = _resolve_cloud_package_name(source, resolved_credentials)
        install_package(package_name)
    else:
        raise ConnectorLoadError("No package source or connector_id specified")

    # Get connector class
    try:
        connector_class = _get_connector_class(package_name)
    except ValueError as e:
        raise ConnectorLoadError(str(e)) from e

    # Build connector kwargs based on source type
    connector_kwargs: dict[str, Any] = {}

    if source.is_cloud:
        required_keys = list(AirbyteCloudAuthConfig.model_fields)
        missing_keys = [k for k in required_keys if k not in resolved_credentials]
        if missing_keys:
            raise ConnectorLoadError(f"Cloud source requires credentials: {required_keys}. Missing: {missing_keys}")

        hosted_class = _get_airbyte_auth_config_class(package_name)
        if hosted_class is None:
            raise ConnectorLoadError(f"Package {package_name} does not support hosted mode (no AirbyteHostedAuthConfig)")

        connector_kwargs["auth_config"] = hosted_class(
            airbyte_client_id=resolved_credentials["airbyte_client_id"],
            airbyte_client_secret=resolved_credentials["airbyte_client_secret"],
            connector_id=source.connector_id,
        )
    else:
        auth_config = _find_matching_auth_config(package_name, resolved_credentials)
        connector_kwargs["auth_config"] = auth_config

    # Add any additional config params
    for key, value in config.config.items():
        connector_kwargs[key] = value

    try:
        return connector_class(**connector_kwargs)
    except Exception as e:
        raise ConnectorLoadError(f"Failed to instantiate connector: {e}") from e



"""Tests for agent helper functions, StreamConsumer, and AgentRunner."""

from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from pydantic_ai import (
    AgentRunResultEvent,
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    PartDeltaEvent,
    PartStartEvent,
    TextPartDelta,
)
from pydantic_ai.messages import RetryPromptPart, TextPart, ToolCallPart, ToolReturnPart

from airbyte_agent_mcp.agent import (
    AgentRunner,
    StreamConsumer,
    _parse_args,
    build_system_prompt,
    create_agent,
    entity_name,
    tool_title,
)


class TestEntityName:
    def test_entity_name_key(self):
        assert entity_name({"entity_name": "users"}) == "users"

    def test_entity_key(self):
        assert entity_name({"entity": "calls"}) == "calls"

    def test_name_key(self):
        assert entity_name({"name": "deals"}) == "deals"

    def test_fallback_to_unknown(self):
        assert entity_name({}) == "unknown"

    def test_priority_order(self):
        assert entity_name({"entity_name": "first", "entity": "second", "name": "third"}) == "first"

    def test_skips_falsy_entity_name(self):
        assert entity_name({"entity_name": "", "entity": "fallback"}) == "fallback"


class TestToolTitle:
    def test_execute_with_entity_and_action(self):
        assert tool_title("execute", {"entity": "users", "action": "list"}) == "Execute (users, list)"

    def test_execute_with_entity_only(self):
        assert tool_title("execute", {"entity": "users"}) == "Execute (users)"

    def test_execute_without_entity(self):
        assert tool_title("execute", {}) == "Execute"

    def test_non_execute_with_underscores(self):
        assert tool_title("get_instructions", {}) == "Get Instructions"

    def test_non_execute_with_hyphens(self):
        assert tool_title("entity-schema", {}) == "Entity Schema"

    def test_simple_name(self):
        assert tool_title("connector_info", {}) == "Connector Info"


class TestBuildSystemPrompt:
    def test_includes_connector_name_and_version(self):
        connector = MagicMock()
        connector.connector_name = "Gong"
        connector.connector_version = "1.2.3"
        connector.list_entities.return_value = []

        prompt = build_system_prompt(connector)
        assert "Gong" in prompt
        assert "1.2.3" in prompt

    def test_includes_entities_with_actions(self):
        connector = MagicMock()
        connector.connector_name = "Test"
        connector.connector_version = "0.1"
        connector.list_entities.return_value = [
            {"entity_name": "users", "available_actions": [{"name": "list"}, {"name": "get"}]},
            {"entity_name": "calls", "available_actions": [{"name": "search"}]},
        ]

        prompt = build_system_prompt(connector)
        assert "users: list, get" in prompt
        assert "calls: search" in prompt

    def test_handles_string_actions(self):
        connector = MagicMock()
        connector.connector_name = "Test"
        connector.connector_version = "0.1"
        connector.list_entities.return_value = [
            {"entity_name": "items", "actions": ["list", "get"]},
        ]

        prompt = build_system_prompt(connector)
        assert "items: list, get" in prompt

    def test_embeds_critical_rules(self):
        connector = MagicMock()
        connector.connector_name = "Test"
        connector.connector_version = "0.1"
        connector.list_entities.return_value = []

        prompt = build_system_prompt(connector)
        assert "current_datetime" in prompt
        assert "CROSS-ENTITY RESOLUTION" in prompt
        assert "entity_schema" in prompt
        assert "select_fields" in prompt
        assert "PAGINATION" in prompt

    def test_does_not_require_get_instructions_first(self):
        connector = MagicMock()
        connector.connector_name = "Test"
        connector.connector_version = "0.1"
        connector.list_entities.return_value = []

        prompt = build_system_prompt(connector)
        assert "Before your first query, call the get_instructions" not in prompt


class TestCreateAgent:
    @patch("airbyte_agent_mcp.agent.Agent")
    @patch("airbyte_agent_mcp.agent.FastMCPToolset")
    def test_creates_agent_with_correct_model(self, mock_toolset_cls, mock_agent_cls):
        connector = MagicMock()
        connector.connector_name = "Test"
        connector.connector_version = "1.0"
        connector.list_entities.return_value = []

        create_agent(connector, model="claude-opus-4-20250514")

        mock_agent_cls.assert_called_once()
        call_args = mock_agent_cls.call_args
        assert call_args[0][0] == "anthropic:claude-opus-4-20250514"

    @patch("airbyte_agent_mcp.agent.Agent")
    @patch("airbyte_agent_mcp.agent.FastMCPToolset")
    def test_passes_system_prompt_as_instructions(self, mock_toolset_cls, mock_agent_cls):
        connector = MagicMock()
        connector.connector_name = "Gong"
        connector.connector_version = "2.0"
        connector.list_entities.return_value = []

        create_agent(connector)

        call_kwargs = mock_agent_cls.call_args[1]
        assert "Gong" in call_kwargs["instructions"]
        assert "2.0" in call_kwargs["instructions"]

    @patch("airbyte_agent_mcp.agent.Agent")
    @patch("airbyte_agent_mcp.agent.FastMCPToolset")
    def test_passes_fastmcp_toolset(self, mock_toolset_cls, mock_agent_cls):
        connector = MagicMock()
        connector.connector_name = "Test"
        connector.connector_version = "1.0"
        connector.list_entities.return_value = []

        create_agent(connector)

        mock_toolset_cls.assert_called_once()
        call_kwargs = mock_agent_cls.call_args[1]
        assert len(call_kwargs["toolsets"]) == 1
        assert call_kwargs["toolsets"][0] == mock_toolset_cls.return_value


class TestParseArgs:
    def test_dict_passthrough(self):
        assert _parse_args({"a": 1}) == {"a": 1}

    def test_json_string(self):
        assert _parse_args('{"entity": "users"}') == {"entity": "users"}

    def test_invalid_json(self):
        assert _parse_args("not json") == {}

    def test_none(self):
        assert _parse_args(None) == {}

    def test_non_dict_json(self):
        assert _parse_args("[1,2,3]") == {}


class TestStreamConsumerDefaults:
    """StreamConsumer methods are no-ops by default."""

    @pytest.mark.asyncio
    async def test_on_tool_call_start_is_noop(self):
        consumer = _make_concrete_consumer()
        await consumer.on_tool_call_start("tool", {}, "id")

    @pytest.mark.asyncio
    async def test_on_tool_call_end_is_noop(self):
        consumer = _make_concrete_consumer()
        await consumer.on_tool_call_end("tool", {}, "result", "id", is_error=False)

    @pytest.mark.asyncio
    async def test_on_intermediate_response_is_noop(self):
        consumer = _make_concrete_consumer()
        await consumer.on_intermediate_response("text")

    @pytest.mark.asyncio
    async def test_on_final_response_is_noop(self):
        consumer = _make_concrete_consumer()
        await consumer.on_final_response("text")

    @pytest.mark.asyncio
    async def test_on_error_is_noop(self):
        consumer = _make_concrete_consumer()
        await consumer.on_error("something broke")


def _make_concrete_consumer() -> StreamConsumer:
    """Create a StreamConsumer with default no-op implementations."""

    class ConcreteConsumer(StreamConsumer): ...

    return ConcreteConsumer()


# --- Helpers for AgentRunner tests ---


def _tool_call_event(tool_name: str, args: dict[str, Any], tool_call_id: str) -> FunctionToolCallEvent:
    return FunctionToolCallEvent(part=ToolCallPart(tool_name=tool_name, args=args, tool_call_id=tool_call_id))


def _tool_result_event(tool_name: str, content: str, tool_call_id: str) -> FunctionToolResultEvent:
    return FunctionToolResultEvent(result=ToolReturnPart(tool_name=tool_name, content=content, tool_call_id=tool_call_id))


def _tool_error_event(tool_name: str, error: str, tool_call_id: str) -> FunctionToolResultEvent:
    return FunctionToolResultEvent(result=RetryPromptPart(content=error, tool_name=tool_name, tool_call_id=tool_call_id))


def _text_start_event(text: str) -> PartStartEvent:
    return PartStartEvent(index=0, part=TextPart(content=text))


def _text_delta_event(text: str) -> PartDeltaEvent:
    return PartDeltaEvent(index=0, delta=TextPartDelta(content_delta=text))


def _run_result_event(messages: list | None = None) -> AgentRunResultEvent:
    mock_result = MagicMock()
    mock_result.all_messages.return_value = messages or []
    return AgentRunResultEvent(result=mock_result)


class RecordingConsumer(StreamConsumer):
    """Records all callbacks for assertions."""

    def __init__(self) -> None:
        self.events: list[tuple[str, dict[str, Any]]] = []

    async def on_tool_call_start(self, tool_name: str, args: dict[str, Any], tool_call_id: str) -> None:
        self.events.append(("tool_call_start", {"tool_name": tool_name, "args": args, "tool_call_id": tool_call_id}))

    async def on_tool_call_end(self, tool_name: str, args: dict[str, Any], result: str, tool_call_id: str, *, is_error: bool) -> None:
        self.events.append(
            (
                "tool_call_end",
                {"tool_name": tool_name, "args": args, "result": result, "tool_call_id": tool_call_id, "is_error": is_error},
            )
        )

    async def on_intermediate_response(self, text: str) -> None:
        self.events.append(("intermediate_response", {"text": text}))

    async def on_final_response(self, text: str) -> None:
        self.events.append(("final_response", {"text": text}))

    async def on_error(self, error: str) -> None:
        self.events.append(("error", {"error": error}))


def _make_runner(events: list, *, raise_after: Exception | None = None) -> AgentRunner:
    """Create an AgentRunner with a mocked agent that yields the given events."""
    agent = MagicMock()

    async def fake_stream(prompt, message_history=None):
        for e in events:
            yield e
        if raise_after:
            raise raise_after

    agent.run_stream_events = fake_stream
    return AgentRunner(agent)


async def _collect(runner: AgentRunner, prompt: str, consumer: RecordingConsumer) -> list:
    """Drain the runner generator and return consumer events."""
    async for _ in runner.run(prompt, consumer):
        pass
    return consumer.events


class TestAgentRunner:
    @pytest.mark.asyncio
    async def test_final_response_text_only(self):
        events = [
            _text_start_event("Hello "),
            _text_delta_event("world"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "hi", consumer)

        assert len(result) == 1
        assert result[0] == ("final_response", {"text": "Hello world"})

    @pytest.mark.asyncio
    async def test_tool_call_dispatches_start_and_end(self):
        events = [
            _tool_call_event("execute", {"entity": "users"}, "tc1"),
            _tool_result_event("execute", '{"data": []}', "tc1"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "list users", consumer)

        assert result[0] == ("tool_call_start", {"tool_name": "execute", "args": {"entity": "users"}, "tool_call_id": "tc1"})
        assert result[1][0] == "tool_call_end"
        assert result[1][1]["tool_name"] == "execute"
        assert result[1][1]["is_error"] is False
        assert result[1][1]["result"] == '{"data": []}'

    @pytest.mark.asyncio
    async def test_tool_error_sets_is_error_flag(self):
        events = [
            _tool_call_event("execute", {}, "tc1"),
            _tool_error_event("execute", "validation failed", "tc1"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "bad call", consumer)

        end_event = next(e for e in result if e[0] == "tool_call_end")
        assert end_event[1]["is_error"] is True
        assert "validation failed" in end_event[1]["result"]

    @pytest.mark.asyncio
    async def test_intermediate_response_flushed_before_tool_call(self):
        events = [
            _text_start_event("thinking... "),
            _tool_call_event("execute", {}, "tc1"),
            _tool_result_event("execute", "ok", "tc1"),
            _text_start_event("done"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "think then act", consumer)

        assert result[0] == ("intermediate_response", {"text": "thinking... "})
        assert result[1][0] == "tool_call_start"
        assert result[-1] == ("final_response", {"text": "done"})

    @pytest.mark.asyncio
    async def test_message_history_accumulates(self):
        messages_run1 = [MagicMock(), MagicMock()]
        events = [
            _text_start_event("response"),
            _run_result_event(messages=messages_run1),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        await _collect(runner, "first", consumer)

        assert runner._message_history == messages_run1

    @pytest.mark.asyncio
    async def test_yields_after_each_callback(self):
        events = [
            _tool_call_event("tool", {}, "tc1"),
            _tool_result_event("tool", "ok", "tc1"),
            _text_delta_event("done"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()

        yield_count = 0
        async for _ in runner.run("test", consumer):
            yield_count += 1

        # tool_call_start, tool_call_end, final_response
        assert yield_count == 3

    @pytest.mark.asyncio
    async def test_non_text_delta_ignored(self):
        """PartDeltaEvent with non-TextPartDelta delta is ignored."""
        non_text_delta = MagicMock()
        non_text_delta.__class__ = type("ThinkingPartDelta", (), {})

        events = [
            PartDeltaEvent(index=0, delta=non_text_delta),
            _text_delta_event("answer"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "think", consumer)

        assert len(result) == 1
        assert result[0] == ("final_response", {"text": "answer"})

    @pytest.mark.asyncio
    async def test_tool_call_args_as_json_string(self):
        """Args provided as JSON string are parsed correctly."""
        events = [
            FunctionToolCallEvent(part=ToolCallPart(tool_name="execute", args='{"entity": "calls"}', tool_call_id="tc1")),
            _tool_result_event("execute", "ok", "tc1"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "test", consumer)

        assert result[0][1]["args"] == {"entity": "calls"}

    @pytest.mark.asyncio
    async def test_empty_text_buffer_no_final_response(self):
        """If no text is emitted, on_final_response is not called."""
        events = [
            _tool_call_event("tool", {}, "tc1"),
            _tool_result_event("tool", "ok", "tc1"),
            _run_result_event(),
        ]
        runner = _make_runner(events)
        consumer = RecordingConsumer()
        result = await _collect(runner, "test", consumer)

        assert not any(e[0] == "final_response" for e in result)
        assert not any(e[0] == "intermediate_response" for e in result)

    @pytest.mark.asyncio
    async def test_unexpected_model_behavior_calls_on_error(self):
        """UnexpectedModelBehavior is caught and dispatched via on_error."""
        from pydantic_ai import UnexpectedModelBehavior

        events = [
            _tool_call_event("execute", {}, "tc1"),
            _tool_result_event("execute", "ok", "tc1"),
        ]
        runner = _make_runner(events, raise_after=UnexpectedModelBehavior("Tool 'execute' exceeded max retries count of 1"))
        consumer = RecordingConsumer()
        result = await _collect(runner, "test", consumer)

        error_event = next(e for e in result if e[0] == "error")
        assert "exceeded max retries" in error_event[1]["error"]



"""Tests for chat CLI commands."""

from unittest.mock import MagicMock, patch

from typer.testing import CliRunner

from airbyte_agent_mcp.cli import cli

runner = CliRunner()


class TestChat:
    def test_missing_config_file(self):
        result = runner.invoke(cli, ["chat", "nonexistent.yaml"])
        assert result.exit_code == 1
        assert "Error" in result.output

    @patch("airbyte_agent_mcp.cli.helpers.load_connector")
    @patch("airbyte_agent_mcp.cli.helpers.ConnectorConfig.load")
    @patch("airbyte_agent_mcp.cli.chat_commands.asyncio")
    def test_oneshot_mode(self, mock_asyncio, mock_config_load, mock_load_connector, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector: test\n")

        mock_connector = MagicMock()
        mock_connector.connector_name = "gong"
        mock_connector.connector_version = "1.0.0"
        mock_load_connector.return_value = mock_connector

        result = runner.invoke(cli, ["chat", str(config_file), "What calls happened?"])
        assert result.exit_code == 0
        mock_asyncio.run.assert_called_once()

    @patch("airbyte_agent_mcp.cli.helpers.load_connector")
    @patch("airbyte_agent_mcp.cli.helpers.ConnectorConfig.load")
    @patch("airbyte_agent_mcp.cli.chat_commands.asyncio")
    def test_interactive_mode(self, mock_asyncio, mock_config_load, mock_load_connector, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector: test\n")

        mock_connector = MagicMock()
        mock_connector.connector_name = "gong"
        mock_connector.connector_version = "1.0.0"
        mock_load_connector.return_value = mock_connector

        result = runner.invoke(cli, ["chat", str(config_file)])
        assert result.exit_code == 0
        mock_asyncio.run.assert_called_once()

    @patch("airbyte_agent_mcp.cli.helpers.load_connector")
    @patch("airbyte_agent_mcp.cli.helpers.ConnectorConfig.load")
    def test_connector_load_error(self, mock_config_load, mock_load_connector, tmp_path):
        from airbyte_agent_mcp.connector_utils import ConnectorLoadError

        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector: test\n")

        mock_load_connector.side_effect = ConnectorLoadError("Package not found")

        result = runner.invoke(cli, ["chat", str(config_file)])
        assert result.exit_code == 1
        assert "Connector load error" in result.output

    @patch("airbyte_agent_mcp.cli.helpers.ConnectorConfig.load")
    def test_config_value_error(self, mock_config_load, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector: test\n")

        mock_config_load.side_effect = ValueError("Invalid config format")

        result = runner.invoke(cli, ["chat", str(config_file)])
        assert result.exit_code == 1
        assert "Configuration error" in result.output



"""CLI command implementations."""

import json
from pathlib import Path
from typing import Annotated

import typer
import yaml
from rich.console import Console
from rich.table import Table

from ..airbyte_api import fetch_registry, get_api, registry_lookup
from ..connector_utils import (
    get_additional_connector_params,
    get_auth_config_types,
    get_cloud_auth_config_type,
)
from ..constants import PACKAGE_PREFIXES
from ..installer import ConnectorInstallError, get_package_name, install_package, is_git_url, is_local_path
from ..models.connector_config import (
    ConnectorConfig,
    ConnectorSource,
    auth_config_to_env_template,
    connector_params_to_template,
)


def _connector_name_from_package(package_name: str) -> str:
    """Derive a connector name from a package name.

    Strips the ``airbyte-agent-`` / ``airbyte_agent_`` prefix when present.
    """
    lower = package_name.lower()
    for prefix in PACKAGE_PREFIXES:
        if lower.startswith(prefix):
            return package_name[len(prefix) :]
    return package_name


def _parse_git_url(spec: str) -> dict:
    """Parse a ``git+https://...`` spec into ConnectorSource fields.

    Handles ``git+<url>@<ref>#subdirectory=<subdir>``.
    """
    url = spec.removeprefix("git+")
    result: dict = {}

    if "#subdirectory=" in url:
        url, subdir = url.rsplit("#subdirectory=", 1)
        result["subdirectory"] = subdir

    if "@" in url:
        url, ref = url.rsplit("@", 1)
        result["ref"] = ref

    result["git"] = url
    return result


connectors_app = typer.Typer(help="Manage Airbyte agent connectors", no_args_is_help=True)
internal_app = typer.Typer(help="Internal commands", hidden=True)
connectors_app.add_typer(internal_app, name="internal")


@connectors_app.command("list-oss")
def list_oss_connectors(
    pattern: Annotated[
        str | None,
        typer.Option("--pattern", "-p", help="Filter connectors by name pattern (e.g., 'salesforce')"),
    ] = None,
) -> None:
    """List available Airbyte agent connectors."""
    console = Console()

    with console.status("Fetching connector registry..."):
        connectors = fetch_registry()

    if pattern:
        pattern_lower = pattern.lower()
        connectors = [c for c in connectors if pattern_lower in c.get("connector_name", "").lower()]

    if not connectors:
        console.print("[yellow]No connectors found matching the criteria.[/yellow]")
        raise typer.Exit(0)

    table = Table()
    table.add_column("Name", style="cyan")
    table.add_column("Package", style="dim")
    table.add_column("Version", style="yellow")
    table.add_column("Connector Definition ID", style="dim")

    for c in connectors:
        name = c.get("connector_name", "")
        table.add_row(
            name,
            f"airbyte-agent-{name}" if name else "",
            c.get("latest_version", ""),
            c.get("connector_definition_id", ""),
        )

    console.print(table)
    console.print(f"\n[dim]Total: {len(connectors)} connectors[/dim]")


@connectors_app.command("list-cloud")
def list_cloud_connectors(
    customer_id: Annotated[
        str | None,
        typer.Option("--customer-id", help="Filter by customer ID"),
    ] = None,
    customer: Annotated[
        str | None,
        typer.Option("--customer", "-c", help="Filter by customer name pattern (e.g. 'acme')"),
    ] = None,
) -> None:
    """List configured cloud connectors."""
    console = Console()
    api = get_api()

    with console.status("Fetching cloud connectors..."):
        if customer_id:
            connectors = api.list_connector_sources(customer_id)
            for c in connectors:
                c["_customer_id"] = customer_id
        else:
            customers = api.list_customers()
            if customer:
                pattern_lower = customer.lower()
                customers = [c for c in customers if pattern_lower in c.get("name", "").lower()]
            connectors = []
            for cust in customers:
                for c in api.list_connector_sources(cust["id"]):
                    c["_customer_id"] = cust["id"]
                    c["_customer_name"] = cust.get("name", "")
                    connectors.append(c)
        registry = registry_lookup()

    if not connectors:
        console.print("[yellow]No cloud connectors found.[/yellow]")
        raise typer.Exit(0)

    table = Table()
    table.add_column("ID", style="cyan")
    table.add_column("Name", style="green")
    table.add_column("Connector", style="yellow")
    table.add_column("Customer", style="dim")
    table.add_column("Created", style="dim")

    for c in connectors:
        defn_id = c.get("summarized_source_template", {}).get("actor_definition_id", "")
        reg = registry.get(defn_id, {})
        customer_label = c.get("_customer_name") or c.get("_customer_id", "")
        table.add_row(
            c.get("id", ""),
            c.get("name", ""),
            reg.get("connector_name", ""),
            customer_label,
            c.get("created_at", ""),
        )

    console.print(table)
    console.print(f"\n[dim]Total: {len(connectors)} connectors[/dim]")


@connectors_app.command("configure")
def configure_connector(
    connector_id: Annotated[str | None, typer.Option("--connector-id", "-c", help="Airbyte Cloud connector ID")] = None,
    package: Annotated[
        str | None,
        typer.Option("--package", help="Package to install: PyPI name, local path, or git+https:// URL"),
    ] = None,
    version: Annotated[str | None, typer.Option("--version", "-v", help="Package version (PyPI only)")] = None,
    filename: Annotated[Path | None, typer.Option("--filename", "-f", help="Output file path (auto-generated if not specified)")] = None,
    overwrite: Annotated[bool, typer.Option("--overwrite", "-o", help="Overwrite output file if it already exists")] = False,
) -> None:
    """Configure a connector by installing it and generating a config."""
    console = Console(stderr=True)

    if not connector_id and not package:
        console.print("[red]Error: Must provide at least one of --connector-id or --package[/red]")
        raise typer.Exit(1)

    if version and not package:
        console.print("[red]Error: --version can only be used with --package[/red]")
        raise typer.Exit(1)

    if version and package and is_local_path(package):
        console.print("[red]Error: --version can only be used with PyPI packages[/red]")
        raise typer.Exit(1)

    try:
        # Build the ConnectorSource
        source_kwargs: dict = {}

        if connector_id:
            source_kwargs["connector_id"] = connector_id

        if package:
            if is_local_path(package):
                source_kwargs["path"] = package
            elif is_git_url(package):
                source_kwargs.update(_parse_git_url(package))
            else:
                source_kwargs["package"] = package
                if version:
                    source_kwargs["version"] = version

        source = ConnectorSource(**source_kwargs)

        # Determine auth types and connector name
        if connector_id:
            api = get_api()
            connector_source = api.get_connector(connector_id)
            registry = registry_lookup()
            defn_id = connector_source.get("source_template", {}).get("source_definition_id", "")
            reg = registry.get(defn_id, {})
            connector_name = reg.get("connector_name", "unknown")
            auth_types = [get_cloud_auth_config_type()]
            additional_params: list = []
            console.print(f"[green]✓[/green] Cloud connector configured: {connector_name} ({connector_id})")
        elif package:
            spec = f"{package}=={version}" if version else package
            install_package(spec)
            package_name = get_package_name(spec)
            connector_name = _connector_name_from_package(package_name)
            auth_types = get_auth_config_types(package_name)
            additional_params = get_additional_connector_params(package_name)
            console.print("[green]✓[/green] Connector installed and configured")
        else:
            console.print("[red]Error: No valid source option provided[/red]")
            raise typer.Exit(1)

        # Create connector config with credentials and config templates
        # Cloud auth fields are global (AIRBYTE_CLIENT_ID), not connector-prefixed
        credentials = auth_config_to_env_template(auth_types[0], connector_name if not connector_id else None)
        config = connector_params_to_template(additional_params)
        connector_config = ConnectorConfig(connector=source, credentials=credentials, config=config)

        # Output config
        config_yaml = yaml.safe_dump(
            connector_config.to_dict(),
            default_flow_style=False,
            sort_keys=False,
        )

        # Add alternative auth methods as comments if there are multiple
        if len(auth_types) > 1:
            for auth_type in auth_types[1:]:
                template = auth_config_to_env_template(auth_type, connector_name)
                yaml_str = yaml.safe_dump({"credentials": template}, default_flow_style=False, sort_keys=False)
                lines = ["", "#", "# Connectors can have more than one way to auth.", "#"]
                for line in yaml_str.strip().split("\n"):
                    lines.append(f"# {line}")
                lines.extend(["#", ""])
                config_yaml += "\n".join(lines)

        if not filename:
            if connector_id:
                source_type = "cloud"
            elif package and is_local_path(package):
                source_type = "path"
            elif package and is_git_url(package):
                source_type = "git"
            else:
                source_type = "package"
            filename = Path(f"connector-{connector_name}-{source_type}.yaml")

    except ConnectorInstallError as e:
        console.print(f"[red]Installation error: {e}[/red]")
        raise typer.Exit(1) from None
    except Exception as e:
        console.print(f"[red]Error: {e}[/red]")
        raise typer.Exit(1) from None

    if filename.exists() and not overwrite:
        console.print(f"[red]Error: Output file already exists: {filename} (use -o to overwrite)[/red]")
        raise typer.Exit(1)

    filename.parent.mkdir(parents=True, exist_ok=True)
    filename.write_text(config_yaml)
    console.print(f"[green]✓[/green] Configuration saved to: {filename}")


@internal_app.command("auth-schemas")
def auth_schemas(
    connector_id: Annotated[str | None, typer.Option("--connector-id", help="Airbyte Cloud connector ID")] = None,
    package: Annotated[
        str | None,
        typer.Option("--package", help="Package to install: PyPI name, local path, or git+https:// URL"),
    ] = None,
    version: Annotated[str | None, typer.Option("--version", "-v", help="Package version (PyPI only)")] = None,
) -> None:
    """Get the JSON schema for a connector's auth configuration."""
    console = Console(stderr=True)

    if not connector_id and not package:
        console.print("[red]Error: Must provide at least one of --connector-id or --package[/red]")
        raise typer.Exit(1)

    if version and not package:
        console.print("[red]Error: --version can only be used with --package[/red]")
        raise typer.Exit(1)

    try:
        if connector_id:
            auth_types = [get_cloud_auth_config_type()]
        elif package:
            spec = f"{package}=={version}" if version else package
            install_package(spec)
            pkg_name = get_package_name(spec)
            auth_types = get_auth_config_types(pkg_name)
        else:
            console.print("[red]Error: No valid source option provided[/red]")
            raise typer.Exit(1)

        schemas = [t.model_json_schema() for t in auth_types]
        print(json.dumps(schemas, indent=2))

    except ConnectorInstallError as e:
        console.print(f"[red]Installation error: {e}[/red]")
        raise typer.Exit(1) from None
    except ModuleNotFoundError as e:
        console.print(f"[red]Error: Package not found: {e}[/red]")
        raise typer.Exit(1) from None
    except ValueError as e:
        console.print(f"[red]Error: {e}[/red]")
        raise typer.Exit(1) from None



from pathlib import Path
from typing import Annotated

import typer
from dotenv import load_dotenv
from rich.console import Console

from ..airbyte_api import AirbyteAuthError
from ..connector_utils import ConnectorLoadError
from ..constants import LOGIN_HINT
from ..models.cli_config import Config, get_org_env_path, set_config_dir
from .chat_commands import chat
from .connector_commands import connectors_app
from .login_commands import login
from .mcp_commands import mcp_app
from .org_commands import orgs_app

cli = typer.Typer(add_completion=False, no_args_is_help=True)


@cli.callback()
def main_callback(
    config_dir: Annotated[
        Path | None,
        typer.Option(
            "--config-dir",
            "-d",
            help="Config directory (default: ~/.airbyte_agent_mcp)",
            envvar="AIRBYTE_CONFIG_DIR",
        ),
    ] = None,
    org: Annotated[
        str | None,
        typer.Option(
            "--org",
            help="Organization ID to use (overrides default)",
        ),
    ] = None,
) -> None:
    """Airbyte Agent MCP CLI."""
    if config_dir:
        set_config_dir(config_dir)

    config = Config.load()

    org_id = org or config.default_organization_id
    if org_id:
        load_dotenv(get_org_env_path(org_id))

    load_dotenv(override=True)


# Add commands and subcommand groups
cli.command("login")(login)
cli.command("chat")(chat)
cli.add_typer(connectors_app, name="connectors")
cli.add_typer(mcp_app, name="mcp")
cli.add_typer(orgs_app, name="orgs")


def main() -> None:
    """Entry point that wraps cli() with error handling."""
    try:
        cli()
    except AirbyteAuthError as e:
        console = Console(stderr=True)
        console.print(f"[red]{e}[/red] {LOGIN_HINT}")
        raise SystemExit(1) from None
    except ConnectorLoadError as e:
        console = Console(stderr=True)
        console.print(f"[red]Connector load error: {e}[/red]")
        raise SystemExit(1) from None
    except FileNotFoundError as e:
        console = Console(stderr=True)
        console.print(f"[red]Error: {e}[/red]")
        raise SystemExit(1) from None


if __name__ == "__main__":
    main()



"""Tests for connector_config models and helpers."""

import os
from unittest.mock import patch

import pytest
from pydantic import BaseModel, ValidationError

from airbyte_agent_mcp.models.connector_config import (
    ConnectorConfig,
    ConnectorSource,
    auth_config_to_env_template,
    connector_params_to_template,
    resolve_credentials,
    resolve_env_vars,
)


class TestConnectorSource:
    def test_package_only(self):
        src = ConnectorSource(package="airbyte-agent-gong")
        assert src.package == "airbyte-agent-gong"
        assert src.version is None
        assert src.path is None
        assert src.git is None
        assert src.connector_id is None

    def test_package_with_version(self):
        src = ConnectorSource(package="airbyte-agent-gong", version="1.0.0")
        assert src.package == "airbyte-agent-gong"
        assert src.version == "1.0.0"

    def test_path_only(self):
        src = ConnectorSource(path="/path/to/connector")
        assert src.path == "/path/to/connector"
        assert src.package is None

    def test_relative_path(self):
        src = ConnectorSource(path="./my/connector")
        assert src.path == "./my/connector"

    def test_git_only(self):
        src = ConnectorSource(git="https://github.com/org/repo.git")
        assert src.git == "https://github.com/org/repo.git"

    def test_git_with_ref_and_subdirectory(self):
        src = ConnectorSource(git="https://github.com/org/repo.git", ref="main", subdirectory="gong")
        assert src.ref == "main"
        assert src.subdirectory == "gong"

    def test_connector_id_only(self):
        src = ConnectorSource(connector_id="abc-123-def-456")
        assert src.connector_id == "abc-123-def-456"
        assert src.is_cloud
        assert not src.has_package_source

    def test_connector_id_with_package(self):
        src = ConnectorSource(connector_id="abc-123", package="airbyte-agent-gong", version="2.0")
        assert src.is_cloud
        assert src.has_package_source
        assert src.package == "airbyte-agent-gong"
        assert src.version == "2.0"

    def test_connector_id_with_path(self):
        src = ConnectorSource(connector_id="abc-123", path="../integrations/gong/.generated")
        assert src.is_cloud
        assert src.has_package_source

    def test_mutual_exclusion_package_and_path(self):
        with pytest.raises(ValidationError, match="At most one"):
            ConnectorSource(package="airbyte-agent-gong", path="/some/path")

    def test_mutual_exclusion_package_and_git(self):
        with pytest.raises(ValidationError, match="At most one"):
            ConnectorSource(package="airbyte-agent-gong", git="https://github.com/org/repo.git")

    def test_mutual_exclusion_path_and_git(self):
        with pytest.raises(ValidationError, match="At most one"):
            ConnectorSource(path="/some/path", git="https://github.com/org/repo.git")

    def test_no_source_at_all(self):
        with pytest.raises(ValidationError, match="At least one"):
            ConnectorSource()

    def test_version_without_package(self):
        with pytest.raises(ValidationError, match="'version' can only be used with 'package'"):
            ConnectorSource(path="/some/path", version="1.0")

    def test_ref_without_git(self):
        with pytest.raises(ValidationError, match="'ref' can only be used with 'git'"):
            ConnectorSource(package="airbyte-agent-gong", ref="main")

    def test_subdirectory_without_git(self):
        with pytest.raises(ValidationError, match="'subdirectory' can only be used with 'git'"):
            ConnectorSource(package="airbyte-agent-gong", subdirectory="gong")


class TestConnectorSourceProperties:
    def test_is_cloud_true(self):
        assert ConnectorSource(connector_id="abc").is_cloud is True

    def test_is_cloud_false(self):
        assert ConnectorSource(package="pkg").is_cloud is False

    def test_has_package_source_package(self):
        assert ConnectorSource(package="pkg").has_package_source is True

    def test_has_package_source_path(self):
        assert ConnectorSource(path="/p").has_package_source is True

    def test_has_package_source_git(self):
        assert ConnectorSource(git="https://example.com/repo.git").has_package_source is True

    def test_has_package_source_cloud_only(self):
        assert ConnectorSource(connector_id="abc").has_package_source is False


class TestConnectorSourceToInstallSpec:
    def test_package_spec(self):
        assert ConnectorSource(package="airbyte-agent-gong").to_install_spec() == "airbyte-agent-gong"

    def test_package_with_version_spec(self):
        assert ConnectorSource(package="airbyte-agent-gong", version="1.0").to_install_spec() == "airbyte-agent-gong==1.0"

    def test_path_spec(self):
        assert ConnectorSource(path="/path/to/connector").to_install_spec() == "/path/to/connector"

    def test_git_spec(self):
        assert ConnectorSource(git="https://github.com/org/repo.git").to_install_spec() == "git+https://github.com/org/repo.git"

    def test_git_with_ref_spec(self):
        src = ConnectorSource(git="https://github.com/org/repo.git", ref="main")
        assert src.to_install_spec() == "git+https://github.com/org/repo.git@main"

    def test_git_with_subdirectory_spec(self):
        src = ConnectorSource(git="https://github.com/org/repo.git", subdirectory="gong")
        assert src.to_install_spec() == "git+https://github.com/org/repo.git#subdirectory=gong"

    def test_git_with_ref_and_subdirectory_spec(self):
        src = ConnectorSource(git="https://github.com/org/repo.git", ref="main", subdirectory="gong")
        assert src.to_install_spec() == "git+https://github.com/org/repo.git@main#subdirectory=gong"

    def test_cloud_only_returns_none(self):
        assert ConnectorSource(connector_id="abc").to_install_spec() is None

    def test_cloud_with_package_returns_package_spec(self):
        src = ConnectorSource(connector_id="abc", package="airbyte-agent-gong", version="2.0")
        assert src.to_install_spec() == "airbyte-agent-gong==2.0"


class TestResolveEnvVars:
    def test_replaces_env_var(self):
        with patch.dict(os.environ, {"MY_KEY": "secret123"}):
            assert resolve_env_vars("${env.MY_KEY}") == "secret123"

    def test_replaces_multiple_vars(self):
        with patch.dict(os.environ, {"A": "hello", "B": "world"}):
            assert resolve_env_vars("${env.A} ${env.B}") == "hello world"

    def test_no_placeholders(self):
        assert resolve_env_vars("plain text") == "plain text"

    def test_missing_env_var_raises(self):
        with patch.dict(os.environ, {}, clear=True):
            with pytest.raises(ValueError, match="MISSING_VAR"):
                resolve_env_vars("${env.MISSING_VAR}")


class TestResolveCredentials:
    def test_resolves_all_values(self):
        with patch.dict(os.environ, {"KEY": "val1", "SECRET": "val2"}):
            result = resolve_credentials({"api_key": "${env.KEY}", "api_secret": "${env.SECRET}"})
            assert result == {"api_key": "val1", "api_secret": "val2"}

    def test_plain_values_unchanged(self):
        result = resolve_credentials({"key": "literal_value"})
        assert result == {"key": "literal_value"}


class TestAuthConfigToEnvTemplate:
    def test_generates_template_with_connector_prefix(self):
        class MyAuth(BaseModel):
            access_key: str
            access_key_secret: str

        result = auth_config_to_env_template(MyAuth, "gong")
        assert result == {
            "access_key": "${env.GONG_ACCESS_KEY}",
            "access_key_secret": "${env.GONG_ACCESS_KEY_SECRET}",
        }

    def test_hyphenated_connector_name(self):
        class MyAuth(BaseModel):
            api_key: str

        result = auth_config_to_env_template(MyAuth, "zendesk-support")
        assert result == {
            "api_key": "${env.ZENDESK_SUPPORT_API_KEY}",
        }

    def test_empty_model(self):
        class Empty(BaseModel):
            pass

        assert auth_config_to_env_template(Empty, "gong") == {}


class TestConnectorParamsToTemplate:
    def test_generates_template(self):
        from airbyte_agent_mcp.connector_utils import ConnectorParam

        params = [ConnectorParam("subdomain", "str", True), ConnectorParam("region", "str", False)]
        result = connector_params_to_template(params)
        assert result == {"subdomain": "# TODO: subdomain", "region": "# TODO: region"}

    def test_empty_list(self):
        assert connector_params_to_template([]) == {}


class TestConnectorConfigStringShorthand:
    def test_string_shorthand_auto_prefixes(self):
        cfg = ConnectorConfig.model_validate({"connector": "gong", "credentials": {}})
        assert cfg.connector.package == "airbyte-agent-gong"

    def test_string_shorthand_already_prefixed(self):
        cfg = ConnectorConfig.model_validate({"connector": "airbyte-agent-gong", "credentials": {}})
        assert cfg.connector.package == "airbyte-agent-gong"

    def test_string_shorthand_with_credentials(self):
        cfg = ConnectorConfig.model_validate({"connector": "gong", "credentials": {"key": "val"}})
        assert cfg.connector.package == "airbyte-agent-gong"
        assert cfg.credentials == {"key": "val"}


class TestConnectorConfigToDict:
    def test_excludes_none_and_empty(self):
        cfg = ConnectorConfig(connector=ConnectorSource(package="airbyte-agent-gong"))
        d = cfg.to_dict()
        assert "credentials" not in d
        assert "config" not in d
        assert d["connector"]["package"] == "airbyte-agent-gong"

    def test_includes_non_empty_credentials(self):
        cfg = ConnectorConfig(
            connector=ConnectorSource(connector_id="gong"),
            credentials={"key": "value"},
        )
        d = cfg.to_dict()
        assert d["credentials"] == {"key": "value"}

    def test_cloud_with_path_to_dict(self):
        cfg = ConnectorConfig(
            connector=ConnectorSource(connector_id="abc-123", path="../integrations/gong/.generated"),
        )
        d = cfg.to_dict()
        assert d["connector"]["connector_id"] == "abc-123"
        assert d["connector"]["path"] == "../integrations/gong/.generated"


class TestConnectorConfigLoadSave:
    def test_load_package_from_yaml(self, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector:\n  package: airbyte-agent-gong\ncredentials:\n  api_key: test\n")
        cfg = ConnectorConfig.load(config_file)
        assert cfg.connector.package == "airbyte-agent-gong"
        assert cfg.credentials == {"api_key": "test"}

    def test_load_path_from_yaml(self, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector:\n  path: /path/to/connector\ncredentials:\n  api_key: test\n")
        cfg = ConnectorConfig.load(config_file)
        assert cfg.connector.path == "/path/to/connector"
        assert cfg.connector.package is None

    def test_load_cloud_from_yaml(self, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector:\n  connector_id: abc-123\ncredentials:\n  key: val\n")
        cfg = ConnectorConfig.load(config_file)
        assert cfg.connector.connector_id == "abc-123"
        assert cfg.connector.is_cloud
        assert not cfg.connector.has_package_source

    def test_load_cloud_with_path_from_yaml(self, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector:\n  connector_id: abc-123\n  path: ../integrations/gong/.generated\n")
        cfg = ConnectorConfig.load(config_file)
        assert cfg.connector.connector_id == "abc-123"
        assert cfg.connector.path == "../integrations/gong/.generated"

    def test_load_string_shorthand_from_yaml(self, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("connector: gong\ncredentials:\n  api_key: test\n")
        cfg = ConnectorConfig.load(config_file)
        assert cfg.connector.package == "airbyte-agent-gong"

    def test_load_missing_file_raises(self, tmp_path):
        with pytest.raises(FileNotFoundError):
            ConnectorConfig.load(tmp_path / "nonexistent.yaml")

    def test_save_and_reload(self, tmp_path):
        cfg = ConnectorConfig(
            connector=ConnectorSource(package="airbyte-agent-gong", version="1.0"),
            credentials={"key": "val"},
        )
        path = tmp_path / "out.yaml"
        cfg.save(path)
        assert path.exists()

        reloaded = ConnectorConfig.load(path)
        assert reloaded.connector.package == "airbyte-agent-gong"
        assert reloaded.connector.version == "1.0"
        assert reloaded.credentials == {"key": "val"}

    def test_save_and_reload_path(self, tmp_path):
        cfg = ConnectorConfig(
            connector=ConnectorSource(path="./my/connector"),
            credentials={"key": "val"},
        )
        path = tmp_path / "out.yaml"
        cfg.save(path)
        reloaded = ConnectorConfig.load(path)
        assert reloaded.connector.path == "./my/connector"

    def test_save_and_reload_cloud_with_package(self, tmp_path):
        cfg = ConnectorConfig(
            connector=ConnectorSource(connector_id="abc-123", package="airbyte-agent-gong", version="2.0"),
        )
        path = tmp_path / "out.yaml"
        cfg.save(path)
        reloaded = ConnectorConfig.load(path)
        assert reloaded.connector.connector_id == "abc-123"
        assert reloaded.connector.package == "airbyte-agent-gong"
        assert reloaded.connector.version == "2.0"

    def test_save_creates_parent_dirs(self, tmp_path):
        cfg = ConnectorConfig(connector=ConnectorSource(connector_id="gong"))
        path = tmp_path / "a" / "b" / "config.yaml"
        cfg.save(path)
        assert path.exists()



"""Tests for connector CLI commands."""

from unittest.mock import patch

from typer.testing import CliRunner

from airbyte_agent_mcp.cli import cli
from airbyte_agent_mcp.cli.connector_commands import _connector_name_from_package, _parse_git_url

runner = CliRunner()


class TestConnectorNameFromPackage:
    def test_strips_airbyte_agent_prefix_hyphen(self):
        assert _connector_name_from_package("airbyte-agent-gong") == "gong"

    def test_strips_airbyte_agent_prefix_underscore(self):
        assert _connector_name_from_package("airbyte_agent_gong") == "gong"

    def test_preserves_name_without_prefix(self):
        assert _connector_name_from_package("custom-connector") == "custom-connector"

    def test_case_insensitive_prefix(self):
        assert _connector_name_from_package("Airbyte-Agent-Gong") == "Gong"

    def test_preserves_compound_name(self):
        assert _connector_name_from_package("airbyte-agent-zendesk-support") == "zendesk-support"


class TestParseGitUrl:
    def test_simple_url(self):
        result = _parse_git_url("git+https://github.com/org/repo.git")
        assert result == {"git": "https://github.com/org/repo.git"}

    def test_url_with_ref(self):
        result = _parse_git_url("git+https://github.com/org/repo.git@v1.0")
        assert result == {"git": "https://github.com/org/repo.git", "ref": "v1.0"}

    def test_url_with_subdirectory(self):
        result = _parse_git_url("git+https://github.com/org/repo.git#subdirectory=pkg")
        assert result == {"git": "https://github.com/org/repo.git", "subdirectory": "pkg"}

    def test_url_with_ref_and_subdirectory(self):
        result = _parse_git_url("git+https://github.com/org/repo.git@main#subdirectory=connectors/gong")
        assert result == {
            "git": "https://github.com/org/repo.git",
            "ref": "main",
            "subdirectory": "connectors/gong",
        }


class TestListOss:
    @patch("airbyte_agent_mcp.cli.connector_commands.fetch_registry")
    def test_lists_connectors(self, mock_fetch):
        mock_fetch.return_value = [
            {"connector_name": "gong", "latest_version": "0.1.0", "connector_definition_id": "abc"},
            {"connector_name": "stripe", "latest_version": "0.2.0", "connector_definition_id": "def"},
        ]
        result = runner.invoke(cli, ["connectors", "list-oss"])
        assert result.exit_code == 0
        assert "gong" in result.output
        assert "stripe" in result.output

    @patch("airbyte_agent_mcp.cli.connector_commands.fetch_registry")
    def test_filters_by_pattern(self, mock_fetch):
        mock_fetch.return_value = [
            {"connector_name": "gong", "latest_version": "0.1.0", "connector_definition_id": "abc"},
            {"connector_name": "stripe", "latest_version": "0.2.0", "connector_definition_id": "def"},
        ]
        result = runner.invoke(cli, ["connectors", "list-oss", "--pattern", "gong"])
        assert result.exit_code == 0
        assert "gong" in result.output
        assert "stripe" not in result.output

    @patch("airbyte_agent_mcp.cli.connector_commands.fetch_registry")
    def test_empty_results(self, mock_fetch):
        mock_fetch.return_value = []
        result = runner.invoke(cli, ["connectors", "list-oss"])
        assert result.exit_code == 0
        assert "No connectors found" in result.output

    @patch("airbyte_agent_mcp.cli.connector_commands.fetch_registry")
    def test_pattern_no_match(self, mock_fetch):
        mock_fetch.return_value = [
            {"connector_name": "gong", "latest_version": "0.1.0", "connector_definition_id": "abc"},
        ]
        result = runner.invoke(cli, ["connectors", "list-oss", "-p", "nonexistent"])
        assert result.exit_code == 0
        assert "No connectors found" in result.output


class TestConfigure:
    def test_requires_connector_id_or_package(self):
        result = runner.invoke(cli, ["connectors", "configure"])
        assert result.exit_code == 1
        assert "Must provide at least one" in result.output

    def test_version_requires_package(self):
        result = runner.invoke(cli, ["connectors", "configure", "--connector-id", "abc", "--version", "1.0.0"])
        assert result.exit_code == 1
        assert "--version can only be used with --package" in result.output

    @patch("airbyte_agent_mcp.cli.connector_commands.install_package")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_package_name")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_auth_config_types")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_additional_connector_params")
    def test_package_mode_generates_config(self, mock_params, mock_auth, mock_pkg_name, mock_install, tmp_path):
        from pydantic import BaseModel, Field

        class TestAuth(BaseModel):
            api_key: str = Field(description="API key")

        mock_pkg_name.return_value = "airbyte-agent-gong"
        mock_auth.return_value = [TestAuth]
        mock_params.return_value = []

        output = tmp_path / "connector-gong-package.yaml"
        result = runner.invoke(cli, ["connectors", "configure", "--package", "airbyte-agent-gong", "-f", str(output)])
        assert result.exit_code == 0
        assert output.exists()
        content = output.read_text()
        assert "connector" in content

    @patch("airbyte_agent_mcp.cli.connector_commands.install_package")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_package_name")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_auth_config_types")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_additional_connector_params")
    def test_overwrite_flag(self, mock_params, mock_auth, mock_pkg_name, mock_install, tmp_path):
        from pydantic import BaseModel, Field

        class TestAuth(BaseModel):
            api_key: str = Field(description="API key")

        mock_pkg_name.return_value = "airbyte-agent-gong"
        mock_auth.return_value = [TestAuth]
        mock_params.return_value = []

        output = tmp_path / "config.yaml"
        output.write_text("old content")

        result = runner.invoke(cli, ["connectors", "configure", "--package", "airbyte-agent-gong", "-f", str(output)])
        assert result.exit_code == 1
        assert "already exists" in result.output

        result = runner.invoke(cli, ["connectors", "configure", "--package", "airbyte-agent-gong", "-f", str(output), "--overwrite"])
        assert result.exit_code == 0

    @patch("airbyte_agent_mcp.cli.connector_commands.install_package")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_package_name")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_auth_config_types")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_additional_connector_params")
    def test_auto_generated_filename(self, mock_params, mock_auth, mock_pkg_name, mock_install, tmp_path, monkeypatch):
        from pydantic import BaseModel, Field

        class TestAuth(BaseModel):
            api_key: str = Field(description="API key")

        mock_pkg_name.return_value = "airbyte-agent-gong"
        mock_auth.return_value = [TestAuth]
        mock_params.return_value = []

        monkeypatch.chdir(tmp_path)

        result = runner.invoke(cli, ["connectors", "configure", "--package", "airbyte-agent-gong"])
        assert result.exit_code == 0
        assert (tmp_path / "connector-gong-package.yaml").exists()

    @patch("airbyte_agent_mcp.cli.connector_commands.get_api")
    @patch("airbyte_agent_mcp.cli.connector_commands.registry_lookup")
    @patch("airbyte_agent_mcp.cli.connector_commands.get_cloud_auth_config_type")
    def test_cloud_mode(self, mock_cloud_auth, mock_registry, mock_api, tmp_path, monkeypatch):
        from pydantic import BaseModel, Field

        class CloudAuth(BaseModel):
            airbyte_client_id: str = Field(description="Client ID")
            airbyte_client_secret: str = Field(description="Client secret")

        mock_cloud_auth.return_value = CloudAuth
        mock_api.return_value.get_connector.return_value = {
            "source_template": {"source_definition_id": "abc-123"},
        }
        mock_registry.return_value = {
            "abc-123": {"connector_name": "gong"},
        }

        monkeypatch.chdir(tmp_path)

        result = runner.invoke(cli, ["connectors", "configure", "-c", "some-connector-id"])
        assert result.exit_code == 0
        assert (tmp_path / "connector-gong-cloud.yaml").exists()



"""MCP CLI commands."""

from pathlib import Path
from typing import Annotated, Literal

import typer
from rich.console import Console

from ..mcp_server import mcp, register_connector_tools
from .helpers import load_connector_from_config
from .register_commands import register_app

mcp_app = typer.Typer(help="MCP server commands", no_args_is_help=True)
mcp_app.add_typer(register_app, name="add-to")


@mcp_app.command("serve")
def serve(
    config: Annotated[
        Path,
        typer.Argument(help="Path to connector config YAML file"),
    ],
    transport: Annotated[
        Literal["stdio", "http"],
        typer.Option("--transport", "-t", help="Transport protocol"),
    ] = "stdio",
    host: Annotated[
        str,
        typer.Option("--host", "-h", help="Host to bind to (for http)"),
    ] = "127.0.0.1",
    port: Annotated[
        int,
        typer.Option("--port", "-p", help="Port to bind to (for http)"),
    ] = 8000,
) -> None:
    """Start the MCP server with a connector configuration."""
    console = Console(stderr=True)
    connector = load_connector_from_config(config, console)
    register_connector_tools(connector)

    if transport == "stdio":
        mcp.run(transport="stdio")
    elif transport == "http":
        mcp.run(transport="http", host=host, port=port)



"""Shared agent setup for pydantic-ai with FastMCP toolset."""

from __future__ import annotations

import json
from collections.abc import AsyncGenerator
from typing import Any, Protocol, runtime_checkable

from pydantic_ai import (
    Agent,
    AgentRunResultEvent,
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    PartDeltaEvent,
    PartStartEvent,
    TextPartDelta,
    UnexpectedModelBehavior,
)
from pydantic_ai.messages import ModelMessage, RetryPromptPart, TextPart
from pydantic_ai.toolsets.fastmcp import FastMCPToolset

from .mcp_server import mcp


def entity_name(entity: dict[str, Any]) -> str:
    """Extract the display name from an entity dict."""
    return entity.get("entity_name") or entity.get("entity") or entity.get("name", "unknown")


def tool_title(name: str, tool_input: dict[str, Any]) -> str:
    """Build a human-readable title for a tool call."""
    if name == "execute" and "entity" in tool_input:
        entity = tool_input["entity"]
        action = tool_input.get("action", "")
        detail = f"{entity}, {action}" if action else entity
        return f"Execute ({detail})"
    return name.replace("_", " ").replace("-", " ").title()


def build_system_prompt(connector: Any) -> str:
    """Build a system prompt that includes connector metadata."""
    lines = []
    for ent in connector.list_entities():
        actions = ent.get("available_actions") or ent.get("actions", [])
        action_names = [a.get("name", str(a)) if isinstance(a, dict) else str(a) for a in actions]
        lines.append(f"  - {entity_name(ent)}: {', '.join(action_names)}")

    return (
        f"You are a helpful data assistant connected to the {connector.connector_name} connector "
        f"(v{connector.connector_version}).\n\n"
        f"Available entities and their actions:\n" + "\n".join(lines) + "\n\n"
        "Use the provided tools to fetch data and answer the user's questions. "
        "When presenting data, format it clearly. "
        "If a query returns many results, summarize the key information. "
        "Always use the tools to get real data rather than making up answers.\n\n"
        "CRITICAL RULES — follow these on every query:\n\n"
        "1. QUERY PLANNING: For any time-based question, call `current_datetime` FIRST. "
        "Date mappings: 'recent'/'last week' = 7 days, 'this week' = since Monday, "
        "'this month' = since the 1st, 'today' = since midnight UTC.\n\n"
        "2. CROSS-ENTITY RESOLUTION: When a question mentions an entity by name, resolve it to an ID "
        "before querying. Use `entity_schema` to find fields ending in `Id`/`_id`, search the related "
        "entity by name, then re-query with the ID as a filter. NEVER paginate through large result "
        "sets to find matches — resolve the ID and filter instead.\n\n"
        "3. ACTION SELECTION: Always use `search` as your default action, not `list`. "
        "Only use `list` when you need today's data (search index may lag) or search returned no results.\n\n"
        "4. FIELD SELECTION: You MUST use `select_fields` or `exclude_fields` on every `execute` call. "
        "Fewer fields = faster and cheaper.\n\n"
        "5. PAGINATION: Use limit of 20-25. Do NOT paginate by default — the first page usually suffices. "
        "For 'how many' questions, say 'at least N' rather than exhaustively paginating. "
        "Never fetch more than 3 pages without checking if you have enough data.\n\n"
        "For the full reference (filter syntax, date range handling, downloads), "
        "call the `get_instructions` tool."
    )


def create_agent(connector: Any, model: str = "claude-opus-4-20250514") -> Agent:
    """Create a pydantic-ai Agent wired to the MCP server's tools.

    Args:
        connector: Instantiated connector object (used for system prompt).
        model: Model identifier (e.g. "claude-opus-4-20250514").

    Returns:
        A configured pydantic-ai Agent.
    """
    return Agent(
        f"anthropic:{model}",
        instructions=build_system_prompt(connector),
        toolsets=[FastMCPToolset(mcp)],
    )


def _parse_args(raw: Any) -> dict[str, Any]:
    """Normalise tool-call args to a dict.

    pydantic-ai may provide args as a dict *or* a JSON string depending on
    the model provider.
    """
    if isinstance(raw, dict):
        return raw
    if isinstance(raw, str):
        try:
            parsed = json.loads(raw)
            if isinstance(parsed, dict):
                return parsed
        except (json.JSONDecodeError, TypeError):
            pass
    return {}


@runtime_checkable
class StreamConsumer(Protocol):
    """Callback interface for consuming agent stream events.

    All methods have default no-op implementations so subclasses only need to
    override the callbacks they care about.
    """

    async def on_tool_call_start(self, tool_name: str, args: dict[str, Any], tool_call_id: str) -> None:
        """Called when the agent invokes a tool."""

    async def on_tool_call_end(self, tool_name: str, args: dict[str, Any], result: str, tool_call_id: str, *, is_error: bool) -> None:
        """Called when a tool call completes (successfully or with an error)."""

    async def on_intermediate_response(self, text: str) -> None:
        """Called when the agent emits text between tool calls."""

    async def on_final_response(self, text: str) -> None:
        """Called once with the final text response after the last agent step."""

    async def on_error(self, error: str) -> None:
        """Called when the agent run fails with an unrecoverable error."""


class AgentRunner:
    """Runs a pydantic-ai Agent and dispatches stream events to a StreamConsumer.

    Manages conversation history internally so callers don't need to track it.
    """

    def __init__(self, agent: Agent) -> None:
        self._agent = agent
        self._message_history: list[ModelMessage] = []

    async def run(self, prompt: str, consumer: StreamConsumer) -> AsyncGenerator[None]:
        """Run the agent with *prompt* and dispatch events to *consumer*.

        This is an async generator that yields after each consumer callback,
        allowing callers to react (e.g. re-render UI) after every event.
        """
        text_buffer = ""
        pending_args: dict[str, dict[str, Any]] = {}

        try:
            async for event in self._agent.run_stream_events(prompt, message_history=self._message_history):
                if isinstance(event, FunctionToolCallEvent):
                    if text_buffer:
                        await consumer.on_intermediate_response(text_buffer)
                        yield
                        text_buffer = ""

                    args = _parse_args(event.part.args)
                    pending_args[event.tool_call_id] = args
                    await consumer.on_tool_call_start(event.part.tool_name, args, event.tool_call_id)
                    yield

                elif isinstance(event, FunctionToolResultEvent):
                    is_error = isinstance(event.result, RetryPromptPart)
                    if is_error:
                        result_str = str(event.result.content) if event.result.content else "unknown error"
                    else:
                        content = event.result.content
                        result_str = content if isinstance(content, str) else json.dumps(content, default=str)

                    args = pending_args.pop(event.tool_call_id, {})
                    tool_name = event.result.tool_name or ""
                    await consumer.on_tool_call_end(tool_name, args, result_str, event.tool_call_id, is_error=is_error)
                    yield

                elif isinstance(event, PartStartEvent) and isinstance(event.part, TextPart):
                    text_buffer += event.part.content

                elif isinstance(event, PartDeltaEvent) and isinstance(event.delta, TextPartDelta):
                    text_buffer += event.delta.content_delta

                elif isinstance(event, AgentRunResultEvent):
                    self._message_history = list(event.result.all_messages())
                    if text_buffer:
                        await consumer.on_final_response(text_buffer)
                        yield
                        text_buffer = ""
        except UnexpectedModelBehavior as exc:
            await consumer.on_error(str(exc))
            yield
